from itertools import tee
import sys
from agouti_pkg.eprint import eprint


class Database(object):
    def __init__(self, database, db_name):
        self.database = database
        self.parent_child_relation = []
        self.features_at_1_level = []
        self.features_at_2_level = []
        self.features_at_3_level = []
        self.name = db_name
        self.parent_child_relation = set()
        self.create_parent_child_relation()
        self.find_featuretypes_at_given_level(3)
        self.find_featuretypes_at_given_level(2)
        self.find_featuretypes_at_given_level(1)

    def create_parent_child_relation(self):
        """Creates Parent -- Child like relations of Feature types
        """

        parent_child = set()
        try:
            file = open("{}.relations".format(self.name))
            lines = file.readlines()
            for line in lines:
                tab = line.strip().split("\t")
                parent_child.add((tab[0], tab[1]))
            file.close()
        except (IOError, IndexError):
            eprint("ERROR: Couldn't find the file {}.relations -> create the database again!".format(self.name))
            sys.exit()
        self.parent_child_relation = parent_child

    def find_featuretypes_at_given_level(self, level):
        """Returns Feature types at specified level

        Arguments:
            level {int} -- level of the feature (1, 2 or 3)
        """

        pc = self.parent_child_relation
        unzipped_pc = list(zip(*pc))
        featuretypes = set()
        l1, l2, l3 = set(), set(), set()
        for p in unzipped_pc[0]:  # finding features at level 1 -> no parents
            if (p not in unzipped_pc[1]):
                l1.add(p)
        for p, c in pc:
            if p in l1:
                l2.add(c)
        for p, c in pc:
            if p in l2:
                l3.add(c)
        if (level == 1):
            self.features_at_1_level = l1
        elif (level == 2):
            self.features_at_2_level = l2
        else:
            self.features_at_3_level = l3
            

    def get_children(self, parent, ids=False, featuretypes=(), level=None):
        """Get children or children ids of the given 'parent'

        Arguments:
            parent {Feature} -- parent Feature

        Keyword Arguments:
            ids {bool} -- True if ids instead of Features are to be returned\
                          (default: {False})
            featuretypes {tuple} -- children of only these featuretypes\
                                    will be returned (default: {()})
            level {int} -- return children at this level (default: {None})

        Returns:
            [list] -- list of chilldren Features or ids
        """

        children = []
        if len(featuretypes):
            child_iter = self.database.children(parent.id,
                                                featuretype=featuretypes,
                                                level=level)
        else:
            child_iter = self.database.children(parent.id, level=level)
        for child in child_iter:
            if(child.featuretype != parent.featuretype):
                if (not ids):
                    children.append(child)
                else:
                    children.append(child.id)
        return children



# -*- coding: utf-8 -*-

"""Powerful and Lightweight Python Tree Data Structure."""

__version__ = "2.4.3"
__author__ = "c0fec0de"
__author_email__ = "c0fec0de@gmail.com"
__description__ = """Powerful and Lightweight Python Tree Data Structure.."""
__url__ = "https://github.com/c0fec0de/anytree"

from .search import CountError  # noqa
from .search import find  # noqa
from .search import find_by_attr  # noqa
from .search import findall  # noqa
from .search import findall_by_attr  # noqa
from .iterators import LevelOrderGroupIter  # noqa
from .iterators import LevelOrderIter  # noqa
from .iterators import PostOrderIter  # noqa
from .iterators import PreOrderIter  # noqa
from .iterators import ZigZagGroupIter  # noqa
from .node import AnyNode  # noqa
from .node import LoopError  # noqa
from .node import Node  # noqa
from .node import NodeMixin  # noqa
from .node import TreeError  # noqa
from .render import AbstractStyle  # noqa
from .render import AsciiStyle  # noqa
from .render import ContRoundStyle  # noqa
from .render import ContStyle  # noqa
from .render import DoubleStyle  # noqa
from .render import RenderTree  # noqa
from .resolver import ChildResolverError  # noqa
from .resolver import Resolver  # noqa
from .resolver import ResolverError  # noqa
from .walker import WalkError  # noqa
from .walker import Walker  # noqa
from . import util  # noqa


# legacy
LevelGroupOrderIter = LevelOrderGroupIter



from .abstractiter import AbstractIter


class LevelOrderGroupIter(AbstractIter):

    """
    Iterate over tree applying level-order strategy with grouping starting at `node`.

    Return a tuple of nodes for each level. The first tuple contains the
    nodes at level 0 (always `node`). The second tuple contains the nodes at level 1
    (children of `node`). The next level contains the children of the children, and so on.

    >>> from anytree import Node, RenderTree, AsciiStyle, LevelOrderGroupIter
    >>> f = Node("f")
    >>> b = Node("b", parent=f)
    >>> a = Node("a", parent=b)
    >>> d = Node("d", parent=b)
    >>> c = Node("c", parent=d)
    >>> e = Node("e", parent=d)
    >>> g = Node("g", parent=f)
    >>> i = Node("i", parent=g)
    >>> h = Node("h", parent=i)
    >>> print(RenderTree(f, style=AsciiStyle()).by_attr())
    f
    |-- b
    |   |-- a
    |   +-- d
    |       |-- c
    |       +-- e
    +-- g
        +-- i
            +-- h
    >>> [[node.name for node in children] for children in LevelOrderGroupIter(f)]
    [['f'], ['b', 'g'], ['a', 'd', 'i'], ['c', 'e', 'h']]
    >>> [[node.name for node in children] for children in LevelOrderGroupIter(f, maxlevel=3)]
    [['f'], ['b', 'g'], ['a', 'd', 'i']]
    >>> [[node.name for node in children]
    ...  for children in LevelOrderGroupIter(f, filter_=lambda n: n.name not in ('e', 'g'))]
    [['f'], ['b'], ['a', 'd', 'i'], ['c', 'h']]
    >>> [[node.name for node in children]
    ...  for children in LevelOrderGroupIter(f, stop=lambda n: n.name == 'd')]
    [['f'], ['b', 'g'], ['a', 'i'], ['h']]
    """

    @staticmethod
    def _iter(children, filter_, stop, maxlevel):
        level = 1
        while children:
            yield tuple([child for child in children if filter_(child)])
            level += 1
            if AbstractIter._abort_at_level(level, maxlevel):
                break
            children = LevelOrderGroupIter._get_grandchildren(children, stop)

    @staticmethod
    def _get_grandchildren(children, stop):
        next_children = []
        for child in children:
            next_children = next_children + AbstractIter._get_children(child.children, stop)
        return next_children



# Sequence Ontology Synonyms
mRNA_synonyms = tuple(map(lambda x: x.lower(),
                          ["mRNA", "SO:0000234", "INSDC_feature:mRNA",
                           "messenger RNA", "protein_coding_transcript",
                           "transcript"]))
cds_synonyms = tuple(map(lambda x: x.lower(),
                         ["CDS", "SO:0000316", "coding sequence",
                          "coding_sequence", "INSDC_feature:CDS"]))
UTR_synonyms = tuple(
    map(lambda x: x.lower(), ["UTR", "SO:0000203", "untranslated region"]))
three_prime_UTR_synonyms = tuple(map(lambda x: x.lower(),
                                     ["three_prime_UTR", "three prime UTR",
                                      "three prime untranslated region",
                                      "3'UTR", "INSDC_feature:3'UTR",
                                      "SO:0000205"]))
five_prime_UTR_synonyms = tuple(map(lambda x: x.lower(),
                                    ["five_prime_UTR", "5'UTR",
                                     "five_prime_untranslated_region",
                                     "INSDC_feature:5'UTR", "SO:0000204"]))



from agouti.gffutils import parser, feature, helpers, constants

def test_feature_from_line():
    # spaces and tabs should give identical results
    line1 = "chr2L	FlyBase	exon	7529	8116	.	+	.	Name=CG11023:1;Parent=FBtr0300689,FBtr0300690"
    line2 = "chr2L FlyBase exon 7529 8116 . + . Name=CG11023:1;Parent=FBtr0300689,FBtr0300690"
    assert feature.feature_from_line(line1, strict=False, keep_order=True) == \
            feature.feature_from_line(line2, strict=False, keep_order=True)


def test_default_feature():
    # Default Feature is 8 tab-delimited ".", with a trailing tab
    assert str(feature.Feature()) == \
    ".	.	.	.	.	.	.	.	"


def test_attributes_representations():
    # These different ways of supplying attributes should yield identical
    # results:
    s = ".	.	.	.	.	.	.	.	ID=asdf"
    for item in (
        '{"ID": ["asdf"]}',
        dict(ID=["asdf"]),
        "ID=asdf"
    ):
        result = str(feature.Feature(attributes=item))
        assert result == s, result


def test_default_start_stop():
    # Whether start or end is "." or None, attribute should always be None and
    # printing should show "."
    c = ['.', None]
    for i1 in c:
        for i2 in c:
            f = feature.Feature(start=i1, end=i2)
            assert f.start is None
            assert f.end is None
            assert f.stop is None
            assert str(f) == ".	.	.	.	.	.	.	.	", str(f)

    # Make sure zero works (protects against sloppy "if start:")
    f = feature.Feature(start=0, end=0)
    assert f.start == f.end == f.stop == 0
    assert str(f) == ".	.	.	0	0	.	.	.	", str(f)


def test_aliases():
    line = "chr2L	FlyBase	exon	7529	8116	.	+	.	Name=CG11023:1;Parent=FBtr0300689,FBtr0300690"
    f = feature.feature_from_line(line, keep_order=True)
    assert f.chrom == 'chr2L' == f.seqid
    assert f.end == 8116 == f.stop

    f.chrom = 'fake'
    f.stop = 1
    assert f.chrom == 'fake' == f.seqid
    assert f.stop == 1 == f.end


def test_string_representation():
    line = "chr2L	FlyBase	exon	7529	8116	.	+	.	Name=CG11023:1;Parent=FBtr0300689,FBtr0300690"
    f = feature.feature_from_line(line, keep_order=True)
    assert line == str(f), str(f)

    line = "chr2L	FlyBase	exon	7529	8116	.	+	.	Name=CG11023:1;Parent=FBtr0300689,FBtr0300690	some	more	stuff"
    f = feature.feature_from_line(line, keep_order=True)
    assert line == str(f)


def test_pbt_interval_conversion():
    try:
        import pybedtools
    except ImportError:
        return
    line = "chr2L FlyBase exon 7529 8116 . + . Name=CG11023:1;Parent=FBtr0300689,FBtr0300690"
    f = feature.feature_from_line(line, strict=False, keep_order=True)
    pbt = helpers.asinterval(f)
    assert pbt.chrom == f.chrom == f.seqid
    assert pbt.start == f.start -1
    assert pbt.stop == f.stop == f.end
    pn = pbt.name
    fn = f.attributes['Name'][0]
    assert pn == fn, '%s, %s' % (pn, fn)


def test_hash():
    line = "chr2L	FlyBase	exon	7529	8116	.	+	.	Name=CG11023:1;Parent=FBtr0300689,FBtr0300690	some	more	stuff"
    f = feature.feature_from_line(line, keep_order=True)
    assert hash(f) == hash(line)


def test_repr():
    line = "chr2L	FlyBase	exon	7529	8116	.	+	.	Name=CG11023:1;Parent=FBtr0300689,FBtr0300690	some	more	stuff"
    f = feature.feature_from_line(line, keep_order=True)
    print(repr(f))
    print(hex(id(f)))
    assert repr(f) == ("<Feature exon (chr2L:7529-8116[+]) at %s>" % hex(id(f)))

def test_attribute_order():

    # default order is gene_id, transcript_id.  But feature_from_line -- if
    # dialect not provided -- will infer its own dialect.  In this case,
    # transcript_id comes first.
    attributes = 'transcript_id "mRNA1"; gene_id "gene1";'
    a = feature.feature_from_line(
        """
        chr1	.	mRNA	1	100	.	+	.	%s
        """ % attributes, strict=False, keep_order=True)
    a.strict = True
    a.keep_order = True
    assert str(a) == 'chr1	.	mRNA	1	100	.	+	.	transcript_id "mRNA1"; gene_id "gene1";', str(a)

    # ensure that using the default dialect uses the default order (and
    # indidentally converts to GFF3 format)
    orig_dialect = a.dialect
    a.dialect = constants.dialect
    a.keep_order = True
    assert str(a) == 'chr1	.	mRNA	1	100	.	+	.	gene_id=gene1;transcript_id=mRNA1', str(a)

    # adding an attribute shoud always result in that attribute coming last (as
    # long as that attribute is not in the dialect order)
    a['dummy'] = ['asdf']
    a.strict = True
    assert str(a) == 'chr1	.	mRNA	1	100	.	+	.	gene_id=gene1;transcript_id=mRNA1;dummy=asdf', str(a)


def test_unjsonify():
    attributes, dialect = parser._split_keyvals('transcript_id "mRNA1"')
    assert attributes == {'transcript_id': ['mRNA1']}, attributes

    s = helpers._jsonify(attributes)
    assert s == '{"transcript_id":["mRNA1"]}', s

    d = helpers._unjsonify(s, isattributes=True)
    assert d == attributes

class IsolatedTestCase(object):
    """
    Isolated test case for checking that the module-level
    constants.always_return_list works.

    This was needed because having this test as a function caused other tests
    to fail even though constants.always_return_list was put back to its
    original setting.  Apparently nose runs tests concurrently in the same
    namespace or something?  Anyway, these setup/teardowns do the trick.
    """
    def setup(self):
        constants.always_return_list = False

    def teardown(self):
        constants.always_return_list = True

    def test_feature_single_item(self):
        line = "chr2L	FlyBase	exon	7529	8116	.	+	.	Name=CG11023:1;Parent=FBtr0300689,FBtr0300690	some	more	stuff"
        f = feature.feature_from_line(line, keep_order=True)
        assert f['Name'] == ['CG11023:1']




class DictExporter(object):

    def __init__(self, dictcls=dict, attriter=None, childiter=list):
        """
        Tree to dictionary exporter.

        Every node is converted to a dictionary with all instance
        attributes as key-value pairs.
        Child nodes are exported to the children attribute.
        A list of dictionaries.

        Keyword Args:
            dictcls: class used as dictionary. :any:`dict` by default.
            attriter: attribute iterator for sorting and/or filtering.
            childiter: child iterator for sorting and/or filtering.

        >>> from pprint import pprint  # just for nice printing
        >>> from anytree import AnyNode
        >>> from anytree.exporter import DictExporter
        >>> root = AnyNode(a="root")
        >>> s0 = AnyNode(a="sub0", parent=root)
        >>> s0a = AnyNode(a="sub0A", b="foo", parent=s0)
        >>> s0b = AnyNode(a="sub0B", parent=s0)
        >>> s1 = AnyNode(a="sub1", parent=root)

        >>> exporter = DictExporter()
        >>> pprint(exporter.export(root))  # order within dictionary might vary!
        {'a': 'root',
         'children': [{'a': 'sub0',
                       'children': [{'a': 'sub0A', 'b': 'foo'}, {'a': 'sub0B'}]},
                      {'a': 'sub1'}]}

        Pythons dictionary `dict` does not preserve order.
        :any:`collections.OrderedDict` does.
        In this case attributes can be ordered via `attriter`.

        >>> from collections import OrderedDict
        >>> exporter = DictExporter(dictcls=OrderedDict, attriter=sorted)
        >>> pprint(exporter.export(root))
        OrderedDict([('a', 'root'),
                     ('children',
                      [OrderedDict([('a', 'sub0'),
                                    ('children',
                                     [OrderedDict([('a', 'sub0A'), ('b', 'foo')]),
                                      OrderedDict([('a', 'sub0B')])])]),
                       OrderedDict([('a', 'sub1')])])])

        The attribute iterator `attriter` may be used for filtering too.
        For example, just dump attributes named `a`:

        >>> exporter = DictExporter(attriter=lambda attrs: [(k, v) for k, v in attrs if k == "a"])
        >>> pprint(exporter.export(root))
        {'a': 'root',
         'children': [{'a': 'sub0', 'children': [{'a': 'sub0A'}, {'a': 'sub0B'}]},
                      {'a': 'sub1'}]}

        The child iterator `childiter` can be used for sorting and filtering likewise:

        >>> exporter = DictExporter(childiter=lambda children: [child for child in children if "0" in child.a])
        >>> pprint(exporter.export(root))
        {'a': 'root',
         'children': [{'a': 'sub0',
                       'children': [{'a': 'sub0A', 'b': 'foo'}, {'a': 'sub0B'}]}]}
        """
        self.dictcls = dictcls
        self.attriter = attriter
        self.childiter = childiter

    def export(self, node):
        """Export tree starting at `node`."""
        attriter = self.attriter or (lambda attr_values: attr_values)
        return self.__export(node, self.dictcls, attriter, self.childiter)

    def __export(self, node, dictcls, attriter, childiter):
        attr_values = attriter(self._iter_attr_values(node))
        attr_values = DictExporter.__filter_node_internals(attr_values)
        data = dictcls(attr_values)
        children = [self.__export(child, dictcls, attriter, childiter)
                    for child in childiter(node.children)]
        if children:
            data['children'] = children
        return data

    def _iter_attr_values(self, node):
        return node.__dict__.items()

    @staticmethod
    def __filter_node_internals(attr_values):
        for attr, value in attr_values:
            if attr in ("_NodeMixin__parent", "_NodeMixin__children"):
                continue
            yield attr, value



import argparse
from agouti_pkg.sequence_ontology import *
import pickle
from agouti_pkg.eprint import eprint
import sys


def validate_arguments(args, parser):
    """ Additional argument validation. Applicable for
        mutually exclusive arguments, etc.

    Arguments:
        args {argparse.Namespace} -- argparse command-line arguments
        parser {ArgumentParser} - ArgumentParser object
    """

    if (args.transcriptomic and args.level == 1):
        parser.error("--level 1 cannot be combined with --transcriptomic")
    elif (not args.transcriptomic and args.annotate_relative_location):
        parser.error("--annotate_relative_location should be used with --transcriptomic")
    elif ((args.combine and (args.attributes or args.features))):
        parser.error("Use --select_features, --select_attributes OR --combine.")
    elif (args.offset != 0 and not args.transcriptomic):
        parser.error("--offset should be used with --transcriptomic")
    elif (args.sep != "\t" and args.custom == "BED"):
        parser.error("--separator must be used with --custom option")

    return


def parse_arguments(argv):
    """Parse command-line arguments for agouti

    Arguments:
        argv {list} -- list of command-line arguments (sys.argv)

    Returns:
        argparse.Namespace -- parsed arguments
    """

    parser = argparse.ArgumentParser('agouti', description='AGouTI - software for annotation of genomic and transcriptomic intervals',
                                    formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    subprasers = parser.add_subparsers(dest='command')
    create_db = subprasers.add_parser('create_db', help='create the database for agouti')
    create_db.add_argument('-a', '--annotation', type=str, help='input file with the genomic annotation in either GTF or GFF3',
                        required=True, dest='annotation')
    create_db.add_argument('-f', '--format', type=str, help='format of the input file with the genomic annotation',
                        required=True, choices=['GTF', 'GFF3'], dest='format')
    create_db.add_argument('-d', '--db', type=str, help='name for the output database',
                        required=True, dest='database')
    create_db.add_argument('-l', '--low-ram', help='enable low-memory mode of the database creation process (warning: slow!)',
                        action="store_true", dest='save_on_disk')
    create_db.add_argument('-i', '--infer_genes', help='infer gene features. Use only with GTF files that do not have lines describing genes (warning: slow!)',
                        action="store_true", dest='infer_genes')
    create_db.add_argument('-j', '--infer_transcripts', help='infer transcript features. Use only with GTF files that do not have lines describing transcripts (warning: slow!)',
                        action="store_true", dest='infer_transcripts')

    annotate = subprasers.add_parser('annotate', help='run annotation with agouti')
    annotate.add_argument('-i', '--input', type=str,
                        help='input file in BED or another column-based format (see --custom).', required=True,
                        dest='bed')
    annotate.add_argument('-d', '--database', type=str, help='database file created with the agouti create_db run mode',
                        required=True, dest='database')
    annotate.add_argument('-m', '--custom', type=str, help='the input text file is in custom format, other than BED. It should contain columns with information about feature id (id), chromosome (chr), start (s) and end (e) coordinates, and optionally about strand. User should provide the proper column indexes (starting from 1) in order: "id,chr,s,e[,strand]". The index of the strand column is optional. Example use: --custom 1,2,4,5,6 or --custom 1,2,4,5. The field separator used in a text file can be specified using the --separator option',
                        required=False, default="BED", dest='custom')
    annotate.add_argument('-p', '--separator', type=str, help='field separator to be used with the --custom option. Default is "\\t"', required=False, default="\t",
                        dest='sep')
    annotate.add_argument('-b', '--coordinates', type=int, help='indicate the coordinate system used in the input file (BED/CUSTOM). Either 0 (0-based coordinates) or 1 (1-based coordinates). Default is 0',
                        choices=[0, 1], required=False, default=0,
                        dest='first_base_num')
    annotate.add_argument('-n', '--header_lines', type=int, help='the number of header lines in the input file. Default is 0',
                        required=False, default=0, dest='header_lines')
    annotate.add_argument('-t', '--transcriptomic', action="store_true", help='transcriptomic annotation mode. In this mode, transcript IDs from the GTF/GFF3 are expected to be placed in the first column of provided BED file instead of chromosome names. Coordinates in this mode are assumed to reflect positions within the transcript')
    annotate.add_argument('-f', '--select_features', type=str, help='comma-separated list of feature names to be reported, e.g., "mRNA,CDS". Refer to [db_name].database.structure.txt file for a list of valid features. By default, all features are reported',
                        required=False, dest='features')
    annotate.add_argument('-a', '--select_attributes', type=str, help='comma-separated list of attribute names to be reported, e.g., "ID,description". Refer to [db_name].database.structure.txt file for a list of valid attributes. By default, all attributes are reported',
                        required=False, dest='attributes')
    annotate.add_argument('-c', '--combine', type=str, help='list of specific feature-attribute combinations to be reported. The combinations should be specified in the format: feature1-attribute1:attribute2,feature2-attribute1, e.g. "mRNA-ID:description,CDS-ID',
                        required=False, dest='combine')
    annotate.add_argument('-s', '--strand_specific', action="store_true",
                        help='strand-specific search')
    annotate.add_argument('-w', '--completly_within', action="store_true",
                        help='the annotated BED interval must be located entirely within the GTF/GFF3 feature. By default, any overlap is sufficient to trigger annotation')
    annotate.add_argument('-l', '--level', type=int, help='annotate results on a specific level (1 for gene level, 2 for mRNA, tRNA level, etc.). For available levels, refer to the tree-like representation of features in [db_name].database.structure.txt file. Please note that --level 1 cannot be combined with –transcriptomic mode. Default is 2',
                        choices=[1, 2], required=False, default=2,
                        dest='level')
    annotate.add_argument('-o', '--offset', type=int, help=argparse.SUPPRESS,
                        required=False, default=0, dest='offset')
    annotate.add_argument('-r', '--annotate_relative_location', action="store_true",
                        help='annotate the relative location of the interval within the feature. Designed to work with –transcriptomic mode')
    annotate.add_argument('--statistics', action="store_true", help='calculate additional feature statistics. Those will be displayed on the stderr',
                        dest='statistics')
    annotate.add_argument('--stats_only', action="store_true", help='calculate and display only feature statistics. No annotation will be performed.', dest='stats_only')
    if len(sys.argv)==1:
        parser.print_help(sys.stderr)
        sys.exit(1)
    args = parser.parse_args()

    if args.command == "create_db":
        if ((args.infer_genes or args.infer_transcripts) and args.format == "GFF3"):
            parser.error("Use option --infer_genes and/or --infer_transcripts only with the GTF file format")
    elif args.command == "annotate":
        args = parser.parse_args(argv[1:])
        validate_arguments(args, parser)

    return args


def create_attributes_and_features_dict(database, args, featuretypes_from_db):
    """Creates dictionary of features (keys) and attributes (values) provided \
        by the user via the command-line. These combinations of features and \
        attributes will be then reported and annotated.

    Arguments:
        database {modules.database.Database} -- object of Database class
        args {argparse.Namespace} -- command-line arguments
        featuretypes_from_db {list} -- list of feature types from the database

    Returns:
        dict -- dictionary of attributes and features to be annotated
    """

    attributes_and_features = {}

    if args.attributes and args.features:
        for feature in list(filter(None, args.features.strip().split(","))):
            attributes_and_features[feature.lower()] = tuple(
                map(lambda x: x.lower(), list(filter(None, args.attributes.strip().split(",")))))
    elif args.combine:
        combined_str = args.combine.strip().split(",")
        for combination in combined_str:
            c = combination.strip().split("-")
            attributes_and_features[c[0].lower()] = tuple(
                map(lambda x: x.lower(), list(filter(None, c[1].strip().split(":")))))
    else:
        try:
            with open('{}.attributes_and_features.pickle'.format(args.database), 'rb') as handle:
                temp_attributes_and_features = pickle.load(handle)
        except FileNotFoundError:
            eprint("Cannot locate file 'attributes_and_features.pickle', which should be created during database creation")
            sys.exit()
        if args.features:
            features_args_list = list(filter(None, args.features.strip().split(",")))
            for feature, attributes in temp_attributes_and_features.items():
                if feature in features_args_list:
                    attributes_and_features[feature.lower()] = attributes
        elif args.attributes:
            attributes_args_list = list(filter(None, args.attributes.strip().split(",")))
            for feature, attributes in temp_attributes_and_features.items():
                attr_list = []
                for attr in attributes:
                    if attr in attributes_args_list:
                        attr_list.append(attr)
                attr_list = list(filter(None, attr_list))
                if len(attr_list):
                    attributes_and_features[feature.lower()] = attr_list
        else:
            attributes_and_features = temp_attributes_and_features

    to_remove = []

    if (args.transcriptomic):
        for f in attributes_and_features.keys():
            if f in database.features_at_3_level and not (
                    f in cds_synonyms or f in UTR_synonyms
                    or f in three_prime_UTR_synonyms or f
                    in five_prime_UTR_synonyms):
                if (args.combine or args.features):
                    eprint("In your case (--transcriptomic), features at the 3rd level must be of CDS or UTR type")
                    sys.exit()
                else:
                    to_remove.append(f)

    for f in set(to_remove):
        del attributes_and_features[f]
    # whether attributes_and_features.keys() are at annotated level
    validity = False
    for a in attributes_and_features.keys():
        if args.level == 1:
            if a in database.features_at_1_level:
                validity = True
        elif args.level == 2:
            if a in database.features_at_2_level:
                validity = True
    if not validity:
        eprint("In your case, you need to specify at least one feature at annotated level")
        sys.exit()
    return attributes_and_features



# pylint: disable=R0913, R0914, C0301
"""
Fasta file -> Faidx -> Fasta -> FastaRecord -> Sequence
"""

from __future__ import division
import os
import sys
from os.path import getmtime
from agouti_pkg.six import PY2, PY3, string_types, integer_types
from agouti_pkg.six.moves import zip_longest
try:
    from collections import OrderedDict
except ImportError:  #python 2.6
    from ordereddict import OrderedDict
from collections import namedtuple
import re
import string
import warnings
from math import ceil
from threading import Lock

if sys.version_info > (3, ):
    buffer = memoryview

dna_bases = re.compile(r'([ACTGNactgnYRWSKMDVHBXyrwskmdvhbx]+)')

__version__ = '0.5.5.2'


class KeyFunctionError(ValueError):
    """Raised if the key_function argument is invalid."""


class FastaIndexingError(Exception):
    """Raised if we encounter malformed FASTA that prevents indexing."""


class IndexNotFoundError(IOError):
    """Raised if read_fai cannot open the index file."""


class FastaNotFoundError(IOError):
    """Raised if the fasta file cannot be opened."""


class FetchError(IndexError):
    """Raised if a request to fetch a FASTA sequence cannot be fulfilled."""


class BedError(ValueError):
    """Indicates a malformed BED entry."""


class RegionError(Exception):
    # This exception class is currently unused, but has been retained for
    # backwards compatibility.
    """A region error occurred."""


class UnsupportedCompressionFormat(IOError):
    """
    Raised when a FASTA file is given with a recognized but unsupported
    compression extension.
    """


class Sequence(object):
    """
    name = FASTA entry name
    seq = FASTA sequence
    start, end = coordinates of subsequence (optional)
    comp = boolean switch for complement property
    """

    def __init__(self, name='', seq='', start=None, end=None, comp=False):
        self.name = name
        self.seq = seq
        self.start = start
        self.end = end
        self.comp = comp
        assert isinstance(name, string_types)
        assert isinstance(seq, string_types)

    def __getitem__(self, n):
        """ Returns a sliced version of Sequence
        >>> x = Sequence(name='chr1', seq='ATCGTA', start=1, end=6)
        >>> x
        >chr1:1-6
        ATCGTA
        >>> x[:3]
        >chr1:1-3
        ATC
        >>> x[3:]
        >chr1:4-6
        GTA
        >>> x[1:-1]
        >chr1:2-5
        TCGT
        >>> x[::-1]
        >chr1:6-1
        ATGCTA
        >>> x[::-3]
        >chr1
        AC
        >>> x = Sequence(name='chr1', seq='ATCGTA', start=0, end=6)
        >>> x
        >chr1:0-6
        ATCGTA
        >>> x[:3]
        >chr1:0-3
        ATC
        >>> x[3:]
        >chr1:3-6
        GTA
        >>> x[1:-1]
        >chr1:1-5
        TCGT
        >>> x[::-1]
        >chr1:6-0
        ATGCTA
        >>> x[::-3]
        >chr1
        AC
        """
        if self.start is None or self.end is None:
            correction_factor = 0
        elif len(
                self.seq
        ) == abs(self.end - self.start) + 1:  # determine coordinate system
            one_based = True
            correction_factor = -1
        elif len(self.seq) == abs(self.end - self.start):
            one_based = False
            correction_factor = 0
        elif len(self.seq) != abs(self.end - self.start):
            raise ValueError(
                "Coordinates (Sequence.start=%s and Sequence.end=%s) imply a different length than Sequence.seq (len=%s). Did you modify Sequence.seq?"
                % (self.start, self.end, len(self.seq)))

        if isinstance(n, slice):
            slice_start, slice_stop, slice_step = n.indices(len(self))
            if self.start is None or self.end is None:  # there should never be self.start != self.end == None
                start = None
                end = None
                return self.__class__(self.name, self.seq[n], start, end,
                                      self.comp)
            self_end, self_start = (self.end, self.start)
            if abs(slice_step) > 1:
                start = None
                end = None
            elif slice_step == -1:  # flip the coordinates when we reverse
                if slice_stop == -1:
                    slice_stop = 0
                start = self_end - slice_stop
                end = self_start + slice_stop
                #print(locals())
            else:
                start = self_start + slice_start
                end = self_start + slice_stop + correction_factor
            return self.__class__(self.name, self.seq[n], start, end,
                                  self.comp)
        elif isinstance(n, integer_types):
            if n < 0:
                n = len(self) + n
            if self.start:
                return self.__class__(self.name, self.seq[n], self.start + n,
                                      self.start + n, self.comp)
            else:
                return self.__class__(self.name, self.seq[n], self.comp)

    def __str__(self):
        return self.seq

    def __neg__(self):
        """ Returns the reverse compliment of sequence
        >>> x = Sequence(name='chr1', seq='ATCGTA', start=1, end=6)
        >>> x
        >chr1:1-6
        ATCGTA
        >>> y = -x
        >>> y
        >chr1:6-1 (complement)
        TACGAT
        >>> -y
        >chr1:1-6
        ATCGTA
        """
        return self[::-1].complement

    def __repr__(self):
        return '\n'.join([''.join(['>', self.fancy_name]), self.seq])

    def __len__(self):
        """
        >>> len(Sequence('chr1', 'ACT'))
        3
        """
        return len(self.seq)

    @property
    def fancy_name(self):
        """ Return the fancy name for the sequence, including start, end, and complementation.
        >>> x = Sequence(name='chr1', seq='ATCGTA', start=1, end=6, comp=True)
        >>> x.fancy_name
        'chr1:1-6 (complement)'
        """
        name = self.name
        if self.start is not None and self.end is not None:
            name = ':'.join([name, '-'.join([str(self.start), str(self.end)])])
        if self.comp:
            name += ' (complement)'
        return name

    @property
    def long_name(self):
        """ DEPRECATED: Use fancy_name instead.
        Return the fancy name for the sequence, including start, end, and complementation.
        >>> x = Sequence(name='chr1', seq='ATCGTA', start=1, end=6, comp=True)
        >>> x.long_name
        'chr1:1-6 (complement)'
        """
        msg = "The `Sequence.long_name` property is deprecated, and will be removed in future versions. Please use `Sequence.fancy_name` instead."
        warnings.warn(msg, DeprecationWarning, stacklevel=2)
        return self.fancy_name

    @property
    def complement(self):
        """ Returns the compliment of self.
        >>> x = Sequence(name='chr1', seq='ATCGTA')
        >>> x.complement
        >chr1 (complement)
        TAGCAT
        """
        comp = self.__class__(
            self.name, complement(self.seq), start=self.start, end=self.end)
        comp.comp = False if self.comp else True
        return comp

    @property
    def reverse(self):
        """ Returns the reverse of self.
        >>> x = Sequence(name='chr1', seq='ATCGTA')
        >>> x.reverse
        >chr1
        ATGCTA
        """
        return self[::-1]

    @property
    def orientation(self):
        """ get the orientation forward=1, reverse=-1
        >>> x = Sequence(name='chr1', seq='ATCGTA', start=1, end=6)
        >>> x.orientation
        1
        >>> x.complement.orientation is None
        True
        >>> x[::-1].orientation is None
        True
        >>> x = -x
        >>> x.orientation
        -1
        """
        if self.start < self.end and not self.comp:
            return 1
        elif self.start > self.end and self.comp:
            return -1
        else:
            return None

    @property
    def gc(self):
        """ Return the GC content of seq as a float
        >>> x = Sequence(name='chr1', seq='ATCGTA')
        >>> y = round(x.gc, 2)
        >>> y == 0.33
        True
        """
        g = self.seq.count('G')
        g += self.seq.count('g')
        c = self.seq.count('C')
        c += self.seq.count('c')
        return (g + c) / len(self.seq)


class IndexRecord(
        namedtuple('IndexRecord',
                   ['rlen', 'offset', 'lenc', 'lenb', 'bend', 'prev_bend'])):
    __slots__ = ()

    def __getitem__(self, key):
        if type(key) == str:
            return getattr(self, key)
        return tuple.__getitem__(self, key)

    def __str__(self):
        return "{rlen:n}\t{offset:n}\t{lenc:n}\t{lenb:n}\n".format(
            **self._asdict())

    def __len__(self):
        return self.rlen


class Faidx(object):
    """ A python implementation of samtools faidx FASTA indexing """

    def __init__(self,
                 filename,
                 default_seq=None,
                 key_function=lambda x: x,
                 as_raw=False,
                 strict_bounds=False,
                 read_ahead=None,
                 mutable=False,
                 split_char=None,
                 duplicate_action="stop",
                 filt_function=lambda x: True,
                 one_based_attributes=True,
                 read_long_names=False,
                 sequence_always_upper=False,
                 rebuild=True,
                 build_index=True):
        """
        filename: name of fasta file
        key_function: optional callback function which should return a unique
          key for the self.index dictionary when given rname.
        as_raw: optional parameter to specify whether to return sequences as a
          Sequence() object or as a raw string.
          Default: False (i.e. return a Sequence() object).
        """
        self.filename = filename

        if filename.lower().endswith('.bgz') or filename.lower().endswith(
                '.gz'):
            # Only try to import Bio if we actually need the bgzf reader.
            try:
                from Bio import bgzf
                from Bio import __version__ as bgzf_version
                from distutils.version import LooseVersion
                if LooseVersion(bgzf_version) < LooseVersion('1.73'):
                    raise ImportError
            except ImportError:
                raise ImportError(
                    "BioPython >= 1.73 must be installed to read block gzip files.")
            else:
                self._fasta_opener = bgzf.open
                self._bgzf = True
        elif filename.lower().endswith('.bz2') or filename.lower().endswith(
                '.zip'):
            raise UnsupportedCompressionFormat(
                "Compressed FASTA is only supported in BGZF format. Use "
                "bgzip to compresss your FASTA.")
        else:
            self._fasta_opener = open
            self._bgzf = False

        try:
            self.file = self._fasta_opener(filename, 'r+b'
                                           if mutable else 'rb')
        except (ValueError, IOError) as e:
            if str(e).find('BGZF') > -1:
                raise UnsupportedCompressionFormat(
                    "Compressed FASTA is only supported in BGZF format. Use "
                    "the samtools bgzip utility (instead of gzip) to "
                    "compress your FASTA.")
            else:
                raise FastaNotFoundError(
                    "Cannot read FASTA file %s" % filename)

        self.indexname = filename + '.fai'
        self.read_long_names = read_long_names
        self.key_function = key_function
        try:
            key_fn_test = self.key_function(
                "TestingReturnType of_key_function")
            if not isinstance(key_fn_test, string_types):
                raise KeyFunctionError(
                    "key_function argument should return a string, not {0}".
                    format(type(key_fn_test)))
        except Exception as e:
            pass
        self.filt_function = filt_function
        assert duplicate_action in ("stop", "first", "last", "longest",
                                    "shortest", "drop")
        self.duplicate_action = duplicate_action
        self.as_raw = as_raw
        self.default_seq = default_seq
        if self._bgzf and self.default_seq is not None:
            raise FetchError(
                "The default_seq argument is not supported with using BGZF compression. Please decompress your FASTA file and try again."
            )
        if self._bgzf:
            self.strict_bounds = True
        else:
            self.strict_bounds = strict_bounds
        self.split_char = split_char
        self.one_based_attributes = one_based_attributes
        self.sequence_always_upper = sequence_always_upper
        self.index = OrderedDict()
        self.lock = Lock()
        self.buffer = dict((('seq', None), ('name', None), ('start', None),
                            ('end', None)))
        if not read_ahead or isinstance(read_ahead, integer_types):
            self.read_ahead = read_ahead
        elif not isinstance(read_ahead, integer_types):
            raise ValueError("read_ahead value must be int, not {0}".format(
                type(read_ahead)))

        self.mutable = mutable
        with self.lock:  # lock around index generation so only one thread calls method
            try:
                if os.path.exists(self.indexname) and getmtime(
                        self.indexname) >= getmtime(self.filename):
                    self.read_fai()
                elif os.path.exists(self.indexname) and getmtime(
                        self.indexname) < getmtime(
                            self.filename) and not rebuild:
                    self.read_fai()
                    warnings.warn(
                        "Index file {0} is older than FASTA file {1}.".format(
                            self.indexname, self.filename), RuntimeWarning)
                elif build_index:
                    self.build_index()
                    self.read_fai()
                else:
                    self.read_fai()

            except FastaIndexingError:
                os.remove(self.indexname)
                self.file.close()
                raise
            except Exception:
                # Handle potential exceptions other than 'FastaIndexingError'
                self.file.close()
                raise

    def __contains__(self, region):
        if not self.buffer['name']:
            return False
        name, start, end = region
        if self.buffer['name'] == name and self.buffer['start'] <= start and self.buffer['end'] >= end:
            return True
        else:
            return False

    def __repr__(self):
        return 'Faidx("%s")' % (self.filename)

    def _index_as_string(self):
        """ Returns the string representation of the index as iterable """
        for k, v in self.index.items():
            yield '\t'.join([k, str(v)])

    def read_fai(self):
        try:
            with open(self.indexname) as index:
                prev_bend = 0
                drop_keys = []
                for line in index:
                    line = line.rstrip()
                    rname, rlen, offset, lenc, lenb = line.split('\t')
                    rlen, offset, lenc, lenb = map(int,
                                                   (rlen, offset, lenc, lenb))
                    newlines = int(ceil(rlen / lenc) * (lenb - lenc))
                    bend = offset + newlines + rlen
                    rec = IndexRecord(rlen, offset, lenc, lenb, bend,
                                      prev_bend)
                    if self.read_long_names:
                        rname = self._long_name_from_index_record(rec)
                    if self.split_char:
                        rname = filter(self.filt_function,
                                       self.key_function(rname).split(
                                           self.split_char))
                    else:
                        # filter must act on an iterable
                        rname = filter(self.filt_function,
                                       [self.key_function(rname)])
                    for key in rname:  # mdshw5/pyfaidx/issues/64
                        if key in self.index:
                            if self.duplicate_action == "stop":
                                raise ValueError('Duplicate key "%s"' % key)
                            elif self.duplicate_action == "first":
                                continue
                            elif self.duplicate_action == "last":
                                self.index[key] = rec
                            elif self.duplicate_action == "longest":
                                if len(rec) > len(self.index[key]):
                                    self.index[key] = rec
                            elif self.duplicate_action == "shortest":
                                if len(rec) < len(self.index[key]):
                                    self.index[key] = rec
                            elif self.duplicate_action == "drop":
                                if key not in drop_keys:
                                    drop_keys.append(key)
                        else:
                            self.index[key] = rec
                    prev_bend = bend
            for dup in drop_keys:
                self.index.pop(dup, None)
        except IOError:
            raise IndexNotFoundError(
                "Could not read index file %s" % self.indexname)

    def build_index(self):
        try:
            with self._fasta_opener(self.filename, 'rb') as fastafile:
                with open(self.indexname, 'w') as indexfile:
                    rname = None  # reference sequence name
                    offset = 0  # binary offset of end of current line
                    rlen = 0  # reference character length
                    blen = None  # binary line length (includes newline)
                    clen = None  # character line length
                    bad_lines = []  # lines > || < than blen
                    thisoffset = offset
                    valid_entry = False
                    lastline = None
                    for i, line in enumerate(fastafile):
                        line_blen = len(line)
                        line = line.decode()
                        line_clen = len(line.rstrip('\n\r'))
                        lastline = i
                        # write an index line
                        if line[0] == '>':
                            valid_entry = check_bad_lines(
                                rname, bad_lines, i - 1)
                            if valid_entry and i > 0:
                                indexfile.write(
                                    "{0}\t{1:d}\t{2:d}\t{3:d}\t{4:d}\n".format(
                                        rname, rlen, thisoffset, clen, blen))
                            elif not valid_entry:
                                raise FastaIndexingError(
                                    "Line length of fasta"
                                    " file is not "
                                    "consistent! "
                                    "Inconsistent line found in >{0} at "
                                    "line {1:n}.".format(
                                        rname, bad_lines[0][0] + 1))
                            blen = None
                            rlen = 0
                            clen = None
                            bad_lines = []
                            try:  # must catch empty deflines (actually these might be okay: https://github.com/samtools/htslib/pull/258)
                                rname = line.rstrip('\n\r')[1:].split()[
                                    0]  # duplicates are detected with read_fai
                            except IndexError:
                                raise FastaIndexingError(
                                    "Bad sequence name %s at line %s." %
                                    (line.rstrip('\n\r'), str(i)))
                            offset += line_blen
                            thisoffset = fastafile.tell(
                            ) if self._bgzf else offset
                        else:  # check line and advance offset
                            if not blen:
                                blen = line_blen
                            if not clen:
                                clen = line_clen
                            # only one short line should be allowed
                            # before we hit the next header, and it
                            # should be the last line in the entry
                            if line_blen != blen or line_blen == 1:
                                bad_lines.append((i, line_blen))
                            offset += line_blen
                            rlen += line_clen
                    
                    # check that we find at least 1 valid FASTA record
                    if not valid_entry:
                        raise FastaIndexingError(
                            "The FASTA file %s does not contain a valid sequence. "
                            "Check that sequence definition lines start with '>'." % self.filename)

                    # write the final index line, if there is one.
                    if lastline is not None:
                        valid_entry = check_bad_lines(
                            rname, bad_lines, lastline
                        )  # advance index since we're at the end of the file
                        if valid_entry:
                            indexfile.write(
                                "{0:s}\t{1:d}\t{2:d}\t{3:d}\t{4:d}\n".format(
                                    rname, rlen, thisoffset, clen, blen))
                        else:
                            raise FastaIndexingError(
                                "Line length of fasta"
                                " file is not "
                                "consistent! "
                                "Inconsistent line found in >{0} at "
                                "line {1:n}.".format(rname,
                                                     bad_lines[0][0] + 1))
        except (IOError, FastaIndexingError) as e:
            if isinstance(e, IOError):
                raise IOError(
                    "%s may not be writable. Please use Fasta(rebuild=False), Faidx(rebuild=False) or faidx --no-rebuild."
                    % self.indexname)
            elif isinstance(e, FastaIndexingError):
                raise e

    def write_fai(self):
        with self.lock:
            with open(self.indexname, 'w') as outfile:
                for line in self._index_as_string:
                    outfile.write(line)

    def from_buffer(self, start, end):
        i_start = start - self.buffer['start']  # want [0, 1) coordinates from [1, 1] coordinates
        i_end = end - self.buffer['start'] + 1
        return self.buffer['seq'][i_start:i_end]

    def fill_buffer(self, name, start, end):
        try:
            seq = self.from_file(name, start, end)
            self.buffer['seq'] = seq
            self.buffer['start'] = start
            self.buffer['end'] = end
            self.buffer['name'] = name
        except FetchError:
            pass

    def fetch(self, name, start, end):
        if self.read_ahead and not (name, start, end) in self:
            self.fill_buffer(name, start, end + self.read_ahead)

        if (name, start, end) in self:
            seq = self.from_buffer(start, end)
        else:
            seq = self.from_file(name, start, end)

        return self.format_seq(seq, name, start, end)

    def from_file(self, rname, start, end, internals=False):
        """ Fetch the sequence ``[start:end]`` from ``rname`` using 1-based coordinates
        1. Count newlines before start
        2. Count newlines to end
        3. Difference of 1 and 2 is number of newlines in [start:end]
        4. Seek to start position, taking newlines into account
        5. Read to end position, return sequence
        """
        assert start == int(start)
        assert end == int(end)
        try:
            i = self.index[rname]
        except KeyError:
            raise FetchError("Requested rname {0} does not exist! "
                             "Please check your FASTA file.".format(rname))
        start0 = start - 1  # make coordinates [0,1)
        if start0 < 0:
            raise FetchError(
                "Requested start coordinate must be greater than 1.")
        seq_len = end - start0

        # Calculate offset (https://github.com/samtools/htslib/blob/20238f354894775ed22156cdd077bc0d544fa933/faidx.c#L398)
        newlines_before = int(
            (start0 - 1) / i.lenc * (i.lenb - i.lenc)) if start0 > 0 else 0
        newlines_to_end = int(end / i.lenc * (i.lenb - i.lenc))
        newlines_inside = newlines_to_end - newlines_before
        seq_blen = newlines_inside + seq_len
        bstart = i.offset + newlines_before + start0
        if seq_blen < 0 and self.strict_bounds:
            raise FetchError("Requested coordinates start={0:n} end={1:n} are "
                             "invalid.\n".format(start, end))
        elif end > i.rlen and self.strict_bounds:
            raise FetchError("Requested end coordinate {0:n} outside of {1}. "
                             "\n".format(end, rname))

        with self.lock:
            if self._bgzf:  # We can't add to virtual offsets, so we need to read from the beginning of the record and trim the beginning if needed
                self.file.seek(i.offset)
                chunk = start0 + newlines_before + newlines_inside + seq_len
                chunk_seq = self.file.read(chunk).decode()
                seq = chunk_seq[start0 + newlines_before:]
            else:
                self.file.seek(bstart)

                if bstart + seq_blen > i.bend and not self.strict_bounds:
                    seq_blen = i.bend - bstart

                if seq_blen > 0:
                    seq = self.file.read(seq_blen).decode()
                elif seq_blen <= 0 and not self.strict_bounds:
                    seq = ''

        if not internals:
            return seq.replace('\n', '').replace('\r', '')
        else:
            return (seq, locals())

    def format_seq(self, seq, rname, start, end):
        start0 = start - 1
        if len(
                seq
        ) < end - start0 and self.default_seq:  # Pad missing positions with default_seq
            pad_len = end - start0 - len(seq)
            seq = ''.join([seq, pad_len * self.default_seq])
        else:  # Return less than requested range
            end = start0 + len(seq)

        if self.sequence_always_upper:
            seq = seq.upper()

        if not self.one_based_attributes:
            start = start0

        if self.as_raw:
            return seq
        else:
            return Sequence(
                name=rname, start=int(start), end=int(end), seq=seq)

    def to_file(self, rname, start, end, seq):
        """ Write sequence in region from start-end, overwriting current
        contents of the FASTA file. """
        if not self.mutable:
            raise IOError(
                "Write attempted for immutable Faidx instance. Set mutable=True to modify original FASTA."
            )
        file_seq, internals = self.from_file(rname, start, end, internals=True)

        with self.lock:
            if len(seq) != len(file_seq) - internals['newlines_inside']:
                raise IOError(
                    "Specified replacement sequence needs to have the same length as original."
                )
            elif len(seq) == len(file_seq) - internals['newlines_inside']:
                line_len = internals['i'].lenc
                if '\r\n' in file_seq:
                    newline_char = '\r\n'
                elif '\r' in file_seq:
                    newline_char = '\r'
                else:
                    newline_char = '\n'
                self.file.seek(internals['bstart'])
                if internals['newlines_inside'] == 0:
                    self.file.write(seq.encode())
                elif internals['newlines_inside'] > 0:
                    n = 0
                    m = file_seq.index(newline_char)
                    while m < len(seq):
                        self.file.write(''.join([seq[n:m], newline_char]).encode())
                        n = m
                        m += line_len
                    self.file.write(seq[n:].encode())
                    self.file.flush()

    def get_long_name(self, rname):
        """ Return the full sequence defline and description. External method using the self.index """
        index_record = self.index[rname]
        if self._bgzf:
            return self._long_name_from_bgzf(index_record)
        else:
            return self._long_name_from_index_record(index_record)

    def _long_name_from_index_record(self, index_record):
        """ Return the full sequence defline and description. Internal method passing IndexRecord """
        prev_bend = index_record.prev_bend
        defline_end = index_record.offset
        self.file.seek(prev_bend)
        return self.file.read(defline_end - prev_bend).decode()[1:-1]

    def _long_name_from_bgzf(self, index_record):
        """ Return the full sequence defline and description. Internal method passing IndexRecord
        This method is present for compatibility with BGZF files, since we cannot subtract their offsets.
        It may be possible to implement a more efficient method. """
        raise NotImplementedError(
            "FastaRecord.long_name and Fasta(read_long_names=True) "
            "are not supported currently for BGZF compressed files.")
        prev_bend = index_record.prev_bend
        self.file.seek(prev_bend)
        defline = []
        while True:
            chunk = self.file.read(4096).decode()
            defline.append(chunk)
            if '\n' in chunk or '\r' in chunk:
                break
        return ''.join(defline)[1:].split('\n\r')[0]

    def close(self):
        self.__exit__()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.file.close()


class FastaRecord(object):
    __slots__ = ['name', '_fa']

    def __init__(self, name, fa):
        self.name = name
        self._fa = fa

    def __getitem__(self, n):
        """Return sequence from region [start, end)

        Coordinates are 0-based, end-exclusive."""
        try:
            if isinstance(n, slice):
                start, stop, step = n.start, n.stop, n.step
                if start is None:
                    start = 0
                if stop is None:
                    stop = len(self)
                if stop < 0:
                    stop = len(self) + stop
                if start < 0:
                    start = len(self) + start
                return self._fa.get_seq(self.name, start + 1, stop)[::step]

            elif isinstance(n, integer_types):
                if n < 0:
                    n = len(self) + n
                return self._fa.get_seq(self.name, n + 1, n + 1)
        except FetchError:
            raise

    def __iter__(self):
        """ Construct a line-based generator that respects the original line lengths. """
        line_len = self._fa.faidx.index[self.name].lenc
        start = 0
        while True:
            end = start + line_len
            if end < len(self):
                yield self[start:end]
            else:
                yield self[start:]
                return
            start += line_len

    def __reversed__(self):
        """ Reverse line-based generator """
        line_len = self._fa.faidx.index[self.name].lenc
        # have to determine last line length
        last_line = len(self) % line_len
        if last_line == 0:
            last_line = line_len
        end = len(self)
        start = end - last_line
        while True:
            if start > 0:
                yield self[start:end][::-1]
            else:
                yield self[:end][::-1]
                return
            if end == len(self):  # first iteration
                end -= last_line
            else:
                end -= line_len
            start = end - line_len

    def __repr__(self):
        return 'FastaRecord("%s")' % (self.name)

    def __len__(self):
        return self._fa.faidx.index[self.name].rlen

    @property
    def unpadded_len(self):
        """ Returns the length of the contig without 5' and 3' N padding.
        Functions the same as contigNonNSize in Fasta.cpp at
        https://github.com/Illumina/hap.py/blob/master/src/c%2B%2B/lib/tools/Fasta.cpp#L284
        """
        length = len(self)
        stop = False
        for line in iter(self):
            if stop:
                break
            if isinstance(line, Sequence):
                line = line.seq
            for base in line.upper():
                if base == 'N':
                    length -= 1
                else:
                    stop = True
                    break
        stop = False
        for line in reversed(self):
            if stop:
                break
            if isinstance(line, Sequence):
                line = line.seq
            for base in line.upper():
                if base == 'N':
                    length -= 1
                else:
                    stop = True
                    break
        return length

    def __str__(self):
        return str(self[:])

    @property
    def variant_sites(self):
        if isinstance(self._fa, FastaVariant):
            pos = []
            var = self._fa.vcf.fetch(self.name, 0, len(self))
            for site in var:
                if site.is_snp:
                    sample = site.genotype(self._fa.sample)
                    if sample.gt_type in self._fa.gt_type and eval(
                            self._fa.filter):
                        pos.append(site.POS)
            return tuple(pos)
        else:
            raise NotImplementedError(
                "variant_sites() only valid for FastaVariant.")

    @property
    def long_name(self):
        """ Read the actual defline from self._fa.faidx mdshw5/pyfaidx#54 """
        return self._fa.faidx.get_long_name(self.name)

    @property
    def __array_interface__(self):
        """ Implement numpy array interface for issue #139"""
        return {
            'shape': (len(self), ),
            'typestr': '|S1',
            'version': 3,
            'data': buffer(str(self).encode('ascii'))
        }


class MutableFastaRecord(FastaRecord):
    def __init__(self, name, fa):
        super(MutableFastaRecord, self).__init__(name, fa)
        if self._fa.faidx._fasta_opener != open:
            raise UnsupportedCompressionFormat(
                "BGZF compressed FASTA is not supported for MutableFastaRecord. "
                "Please decompress your FASTA file.")

    def __setitem__(self, n, value):
        """Mutate sequence in region [start, end)
        to value.
        Coordinates are 0-based, end-exclusive."""
        try:
            if isinstance(n, slice):
                start, stop, step = n.start, n.stop, n.step
                if step:
                    raise IndexError("Step operator is not implemented.")
                if not start:
                    start = 0
                if not stop:
                    stop = len(self)
                if stop < 0:
                    stop = len(self) + stop
                if start < 0:
                    start = len(self) + start
                self._fa.faidx.to_file(self.name, start + 1, stop, value)

            elif isinstance(n, integer_types):
                if n < 0:
                    n = len(self) + n
                return self._fa.faidx.to_file(self.name, n + 1, n + 1, value)
        except (FetchError, IOError):
            raise


class Fasta(object):
    def __init__(self,
                 filename,
                 default_seq=None,
                 key_function=lambda x: x,
                 as_raw=False,
                 strict_bounds=False,
                 read_ahead=None,
                 mutable=False,
                 split_char=None,
                 filt_function=lambda x: True,
                 one_based_attributes=True,
                 read_long_names=False,
                 duplicate_action="stop",
                 sequence_always_upper=False,
                 rebuild=True,
                 build_index=True):
        """
        An object that provides a pygr compatible interface.
        filename: name of fasta file
        """
        self.filename = filename
        self.mutable = mutable
        self.faidx = Faidx(
            filename,
            key_function=key_function,
            as_raw=as_raw,
            default_seq=default_seq,
            strict_bounds=strict_bounds,
            read_ahead=read_ahead,
            mutable=mutable,
            split_char=split_char,
            filt_function=filt_function,
            one_based_attributes=one_based_attributes,
            read_long_names=read_long_names,
            duplicate_action=duplicate_action,
            sequence_always_upper=sequence_always_upper,
            rebuild=rebuild,
            build_index=build_index)
        self.keys = self.faidx.index.keys
        if not self.mutable:
            self.records = dict(
                [(rname, FastaRecord(rname, self)) for rname in self.keys()])
        elif self.mutable:
            self.records = dict([(rname, MutableFastaRecord(rname, self))
                                 for rname in self.keys()])

    def __contains__(self, rname):
        """Return True if genome contains record."""
        return rname in self.faidx.index

    def __getitem__(self, rname):
        """Return a chromosome by its name, or its numerical index."""
        if isinstance(rname, integer_types):
            rname = tuple(self.keys())[rname]
        try:
            return self.records[rname]
        except KeyError:
            raise KeyError("{0} not in {1}.".format(rname, self.filename))

    def __repr__(self):
        return 'Fasta("%s")' % (self.filename)

    def __iter__(self):
        for rname in self.keys():
            yield self[rname]

    def get_seq(self, name, start, end, rc=False):
        """Return a sequence by record name and interval [start, end).

        Coordinates are 1-based, end-exclusive.
        If rc is set, reverse complement will be returned.
        """
        # Get sequence from real genome object and save result.
        seq = self.faidx.fetch(name, start, end)
        if rc:
            return -seq
        else:
            return seq

    def get_spliced_seq(self, name, intervals, rc=False):
        """Return a sequence by record name and list of intervals

        Interval list is an iterable of [start, end].
        Coordinates are 1-based, end-exclusive.
        If rc is set, reverse complement will be returned.
        """
        # Get sequence for all intervals
        chunks = [self.faidx.fetch(name, s, e) for s, e in intervals]
        start = chunks[0].start
        end = chunks[-1].end

        # reverce complement
        if rc:
            seq = "".join([(-chunk).seq for chunk in chunks[::-1]])
        else:
            seq = "".join([chunk.seq for chunk in chunks])

        # Sequence coordinate validation wont work since
        # len(Sequence.seq) != end - start
        return Sequence(name=name, seq=seq, start=None, end=None)

    def close(self):
        self.__exit__()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.faidx.__exit__(*args)


class FastaVariant(Fasta):
    """ Return consensus sequence from FASTA and VCF inputs
    """
    expr = set(('>', '<', '=', '!'))

    def __init__(self,
                 filename,
                 vcf_file,
                 sample=None,
                 het=True,
                 hom=True,
                 call_filter=None,
                 **kwargs):
        super(FastaVariant, self).__init__(filename, **kwargs)
        try:
            import pysam
        except ImportError:
            raise ImportError("pysam must be installed for FastaVariant.")
        try:
            import vcf
        except ImportError:
            raise ImportError("PyVCF must be installed for FastaVariant.")
        if call_filter is not None:
            try:
                key, expr, value = call_filter.split()  # 'GQ > 30'
            except IndexError:
                raise ValueError(
                    "call_filter must be a string in the format 'XX <>!= NN'")
            assert all([x in self.expr for x in list(expr)])
            assert all([x in string.ascii_uppercase for x in list(key)])
            assert all([x in string.printable for x in list(value)])
            self.filter = "sample['{key}'] {expr} {value}".format(**locals())
        else:
            self.filter = 'True'
        if os.path.exists(vcf_file):
            self.vcf = vcf.Reader(filename=vcf_file)
        else:
            raise IOError("File {0} does not exist.".format(vcf_file))
        if sample is not None:
            self.sample = sample
        else:
            self.sample = self.vcf.samples[0]
            if len(self.vcf.samples) > 1 and sample is None:
                warnings.warn("Using sample {0} genotypes.".format(
                    self.sample), RuntimeWarning)
        if het and hom:
            self.gt_type = set((1, 2))
        elif het:
            self.gt_type = set((1, ))
        elif hom:
            self.gt_type = set((2, ))
        else:
            self.gt_type = set()

    def __repr__(self):
        return 'FastaVariant("%s", "%s", gt="%s")' % (self.filename,
                                                      self.vcf.filename,
                                                      str(self.gt_type))

    def get_seq(self, name, start, end):
        """Return a sequence by record name and interval [start, end).
        Replace positions with polymorphism with variant.
        Coordinates are 0-based, end-exclusive.
        """
        seq = self.faidx.fetch(name, start, end)
        if self.faidx.as_raw:
            seq_mut = list(seq)
            del seq
        else:
            seq_mut = list(seq.seq)
            del seq.seq
        var = self.vcf.fetch(name, start - 1, end)
        for record in var:
            if record.is_snp:  # skip indels
                sample = record.genotype(self.sample)
                if sample.gt_type in self.gt_type and eval(self.filter):
                    alt = record.ALT[0]
                    i = (record.POS - 1) - (start - 1)
                    seq_mut[i:i + len(alt)] = str(alt)
        # slice the list in case we added an MNP in last position
        if self.faidx.as_raw:
            return ''.join(seq_mut[:end - start + 1])
        else:
            seq.seq = ''.join(seq_mut[:end - start + 1])
            return seq


def wrap_sequence(n, sequence, fillvalue=''):
    args = [iter(sequence)] * n
    for line in zip_longest(fillvalue=fillvalue, *args):
        yield ''.join(line + ("\n", ))


# To take a complement, we map each character in the first string in this pair
# to the corresponding character in the second string.
complement_map = ('ACTGNactgnYRWSKMDVHBXyrwskmdvhbx',
                  'TGACNtgacnRYWSMKHBDVXrywsmkhbdvx')
invalid_characters_set = set(
    chr(x) for x in range(256) if chr(x) not in complement_map[0])
invalid_characters_string = ''.join(invalid_characters_set)

if PY3:
    complement_table = str.maketrans(complement_map[0], complement_map[1],
                                     invalid_characters_string)
    translate_arguments = (complement_table, )
elif PY2:
    complement_table = string.maketrans(complement_map[0], complement_map[1])
    translate_arguments = (complement_table, invalid_characters_string)


def complement(seq):
    """ Returns the complement of seq.
    >>> seq = 'ATCGTA'
    >>> complement(seq)
    'TAGCAT'
    """
    seq = str(seq)
    result = seq.translate(*translate_arguments)
    if len(result) != len(seq):
        first_invalid_position = next(
            i for i in range(len(seq)) if seq[i] in invalid_characters_set)
        raise ValueError(
            "Sequence contains non-DNA character '{0}' at position {1:n}\n".
            format(seq[first_invalid_position], first_invalid_position + 1))
    return result


def translate_chr_name(from_name, to_name):
    chr_name_map = dict(zip(from_name, to_name))

    def map_to_function(rname):
        return chr_name_map[rname]

    return map_to_function


def bed_split(bed_entry):
    try:
        rname, start, end = bed_entry.rstrip().split()[:3]
    except (IndexError, ValueError):
        raise BedError('Malformed BED entry! {0}\n'.format(bed_entry.rstrip()))
    start, end = (int(start), int(end))
    return (rname, start, end)


def ucsc_split(region):
    try:
        rname, interval = region.split(':')
    except ValueError:
        rname = region
        interval = None
    try:
        start, end = interval.split('-')
        start, end = (int(start) - 1, int(end))
    except (AttributeError, ValueError):
        start, end = (None, None)
    return (rname, start, end)


def check_bad_lines(rname, bad_lines, i):
    """ Find inconsistent line lengths in the middle of an
    entry. Allow blank lines between entries, and short lines
    occurring at the last line of an entry. Returns boolean
    validating the entry.
    >>> check_bad_lines('chr0', [(10, 79)], 10)
    True
    >>> check_bad_lines('chr0', [(9, 79)], 10)
    False
    >>> check_bad_lines('chr0', [(9, 79), (10, 1)], 10)
    True
    """
    if len(bad_lines) == 0:
        return True
    elif len(bad_lines) == 1:
        if bad_lines[0][0] == i:  # must be last line
            return True
        else:
            return False
    elif len(bad_lines) == 2:
        if bad_lines[0][0] == i:  # must not be last line
            return False
        elif bad_lines[1][0] == i and bad_lines[1][1] == 1:  # blank last line
            if bad_lines[0][0] + 1 == i and bad_lines[0][1] > 1:  # non-blank line
                return True
        else:
            return False
    if len(bad_lines) > 2:
        return False
    raise RuntimeError("Unhandled exception during fasta indexing at entry " + rname + \
                       "Please report this issue at https://github.com/mdshw5/pyfaidx/issues " + \
                       str(bad_lines))


if __name__ == "__main__":
    import doctest
    doctest.testmod()



import agouti_pkg.six
import collections
from agouti_pkg.gffutils import constants


# collections.MutableMapping is apparently the best way to provide dict-like
# interface (http://stackoverflow.com/a/3387975)
try:
    class Attributes(collections.MutableMapping):
        def __init__(self, *args, **kwargs):
            """
            An Attributes object acts much like a dictionary.  However, values are
            always stored internally as lists, even if a single value is provided.

            Whether or not you get a list back depends on the
            `constants.always_return_list` setting, which can be set on-the-fly.
            If True, then one-item lists are returned.  This is best shown with an
            example:

            Set up an Attributes object:

                >>> attr = Attributes()

            Set the "Name" attribute with a string:

                >>> attr['Name'] = 'gene1'

            This is stored internally as a list, and by default, we'll get a list
            back:

                >>> assert attr['Name'] == ['gene1']

            The same thing happens if we set it with a list in the first place:

                >>> attr['Name'] = ['gene1']
                >>> assert attr['Name'] == ['gene1']

            Now, change the setting so that upon access, single-value lists are
            returned as the first item.

                >>> constants.always_return_list = False
                >>> assert attr['Name'] == 'gene1'

            Change it back again:

                >>> constants.always_return_list = True
                >>> assert attr['Name'] == ['gene1']

            """
            self._d = dict()
            self.update(*args, **kwargs)

        def __setitem__(self, k, v):
            if not isinstance(v, (list, tuple)):
                v = [v]
            self._d[k] = v

        def __getitem__(self, k):
            v = self._d[k]
            if constants.always_return_list:
                return v
            if isinstance(v, list) and len(v) == 1:
                v = v[0]
            return v

        def __delitem__(self, key):
            del self._d[key]

        def __iter__(self):
            return iter(self.keys())

        def __len__(self):
            return len(self._d)

        def keys(self):
            return self._d.keys()

        def values(self):
            return [self.__getitem__(k) for k in self.keys()]

        def items(self):
            r = []
            for k in self.keys():
                r.append((k, self.__getitem__(k)))
            return r

        def __str__(self):
            s = []
            for i in self.items():
                s.append("%s: %s" % i)
            return '\n'.join(s)

        def update(self, *args, **kwargs):
            for k, v in agouti_pkg.six.iteritems(dict(*args, **kwargs)):
                self[k] = v
except AttributeError:
    class Attributes(collections.abc.MutableMapping):
        def __init__(self, *args, **kwargs):
            """
            An Attributes object acts much like a dictionary.  However, values are
            always stored internally as lists, even if a single value is provided.

            Whether or not you get a list back depends on the
            `constants.always_return_list` setting, which can be set on-the-fly.
            If True, then one-item lists are returned.  This is best shown with an
            example:

            Set up an Attributes object:

                >>> attr = Attributes()

            Set the "Name" attribute with a string:

                >>> attr['Name'] = 'gene1'

            This is stored internally as a list, and by default, we'll get a list
            back:

                >>> assert attr['Name'] == ['gene1']

            The same thing happens if we set it with a list in the first place:

                >>> attr['Name'] = ['gene1']
                >>> assert attr['Name'] == ['gene1']

            Now, change the setting so that upon access, single-value lists are
            returned as the first item.

                >>> constants.always_return_list = False
                >>> assert attr['Name'] == 'gene1'

            Change it back again:

                >>> constants.always_return_list = True
                >>> assert attr['Name'] == ['gene1']

            """
            self._d = dict()
            self.update(*args, **kwargs)

        def __setitem__(self, k, v):
            if not isinstance(v, (list, tuple)):
                v = [v]
            self._d[k] = v

        def __getitem__(self, k):
            v = self._d[k]
            if constants.always_return_list:
                return v
            if isinstance(v, list) and len(v) == 1:
                v = v[0]
            return v

        def __delitem__(self, key):
            del self._d[key]

        def __iter__(self):
            return iter(self.keys())

        def __len__(self):
            return len(self._d)

        def keys(self):
            return self._d.keys()

        def values(self):
            return [self.__getitem__(k) for k in self.keys()]

        def items(self):
            r = []
            for k in self.keys():
                r.append((k, self.__getitem__(k)))
            return r

        def __str__(self):
            s = []
            for i in self.items():
                s.append("%s: %s" % i)
            return '\n'.join(s)

        def update(self, *args, **kwargs):
            for k, v in agouti_pkg.six.iteritems(dict(*args, **kwargs)):
                self[k] = v
    


# Useful for profiling: which dictionary-like class to store attributes in.
# This is used in Feature below and in parser.py

dict_class = Attributes
#dict_class = dict
#dict_class = helper_classes.DefaultOrderedDict
#dict_class = collections.defaultdict
#dict_class = collections.OrderedDict
#dict_class = helper_classes.DefaultListOrderedDict



import warnings

from anytree.exporter.dotexporter import DotExporter


class RenderTreeGraph(DotExporter):

    def __init__(self, *args, **kwargs):
        """Legacy. Use :any:`anytree.exporter.DotExporter` instead."""
        warnings.warn(
            ("anytree.RenderTreeGraph has moved. "
             "Use anytree.exporter.DotExporter instead"),
            DeprecationWarning
        )
        super(RenderTreeGraph, self).__init__(*args, **kwargs)



"""
Module for integration with pybedtools
"""

import os
import pybedtools
from pybedtools import featurefuncs
from agouti_pkg.gffutils import helpers
import agouti_pkg.six


def to_bedtool(iterator):
    """
    Convert any iterator into a pybedtools.BedTool object.

    Note that the supplied iterator is not consumed by this function. To save
    to a temp file or to a known location, use the `.saveas()` method of the
    returned BedTool object.
    """
    def gen():
        for i in iterator:
            yield helpers.asinterval(i)
    return pybedtools.BedTool(gen())


def tsses(db, merge_overlapping=False, attrs=None, attrs_sep=":",
          merge_kwargs=None, as_bed6=False, bedtools_227_or_later=True):
    """
    Create 1-bp transcription start sites for all transcripts in the database
    and return as a sorted pybedtools.BedTool object pointing to a temporary
    file.

    To save the file to a known location, use the `.moveto()` method on the
    resulting `pybedtools.BedTool` object.

    To extend regions upstream/downstream, see the `.slop()` method on the
    resulting `pybedtools.BedTool object`.

    Requires pybedtools.

    Parameters
    ----------
    db : gffutils.FeatureDB
        The database to use

    as_bed6 : bool
        If True, output file is in BED6 format; otherwise it remains in the
        GFF/GTF format and dialect of the file used to create the database.

        Note that the merge options below necessarily force `as_bed6=True`.

    merge_overlapping : bool
        If True, output will be in BED format. Overlapping TSSes will be merged
        into a single feature, and their names will be collapsed using
        `merge_sep` and placed in the new name field.

    merge_kwargs : dict
        If `merge_overlapping=True`, these keyword arguments are passed to
        pybedtools.BedTool.merge(), which are in turn sent to `bedtools merge`.
        The merge operates on a BED6 file which will have had the name field
        constructed as specified by other arguments here.  See the available
        options for your installed version of BEDTools; the defaults used here
        are `merge_kwargs=dict(o='distinct', c=4, s=True)`.

        Any provided `merge_kwargs` are used to *update* the default. It is
        recommended to not override `c=4` and `s=True`, otherwise the
        post-merge fixing may not work correctly. Good candidates for tweaking
        are `d` (merge distance), `o` (operation), `delim` (delimiter to use
        for collapse operations).

    attrs : str or list
        Only has an effect when `as_bed6=True` or `merge_overlapping=True`.

        Determines what goes in the name field of an output BED file. By
        default, "gene_id" for GTF databases and "ID" for GFF. If a list of
        attributes is supplied, e.g. ["gene_id", "transcript_id"], then these
        will be joined by `attr_join_sep` and then placed in the name field.

    attrs_sep: str
        If `as_bed6=True` or `merge_overlapping=True`, then use this character
        to separate attributes in the name field of the output BED. If also
        using `merge_overlapping=True`, you'll probably want this to be
        different than `merge_sep` in order to parse things out later.

    bedtools_227_or_later : bool
        In version 2.27, BEDTools changed the output for merge. By default,
        this function expects BEDTools version 2.27 or later, but set this to
        False to assume the older behavior.

        For testing purposes, the environment variable
        GFFUTILS_USES_BEDTOOLS_227_OR_LATER is set to either "true" or "false"
        and is used to override this argument.

    Examples
    --------

    >>> import gffutils
    >>> db = gffutils.create_db(
    ...    gffutils.example_filename('FBgn0031208.gtf'),
    ...    ":memory:",
    ...    keep_order=True,
    ...    verbose=False)

    Default settings -- no merging, and report a separate TSS on each line even
    if they overlap (as in the first two):


    >>> print(tsses(db))                        # doctest: +NORMALIZE_WHITESPACE
    chr2L	gffutils_derived	transcript_TSS	7529	7529	.	+	.	gene_id "FBgn0031208"; transcript_id "FBtr0300689";
    chr2L	gffutils_derived	transcript_TSS	7529	7529	.	+	.	gene_id "FBgn0031208"; transcript_id "FBtr0300690";
    chr2L	gffutils_derived	transcript_TSS	11000	11000	.	-	.	gene_id "Fk_gene_1"; transcript_id "transcript_Fk_gene_1";
    chr2L	gffutils_derived	transcript_TSS	12500	12500	.	-	.	gene_id "Fk_gene_2"; transcript_id "transcript_Fk_gene_2";
    <BLANKLINE>


    Default merging, showing the first two TSSes merged and reported as
    a single unique TSS for the gene. Note the conversion to BED:

    >>> x = tsses(db, merge_overlapping=True)
    >>> print(x)  # doctest: +NORMALIZE_WHITESPACE
    chr2L	7528	7529	FBgn0031208	.	+
    chr2L	10999	11000	Fk_gene_1	.	-
    chr2L	12499	12500	Fk_gene_2	.	-
    <BLANKLINE>

    Report both gene ID and transcript ID in the name. In some cases this can
    be easier to parse than the original GTF or GFF file. With no merging
    specified, we must add `as_bed6=True` to see the names in BED format.

    >>> x = tsses(db, attrs=['gene_id', 'transcript_id'], as_bed6=True)
    >>> print(x)  # doctest: +NORMALIZE_WHITESPACE
    chr2L	7528	7529	FBgn0031208:FBtr0300689	.	+
    chr2L	7528	7529	FBgn0031208:FBtr0300690	.	+
    chr2L	10999	11000	Fk_gene_1:transcript_Fk_gene_1	.	-
    chr2L	12499	12500	Fk_gene_2:transcript_Fk_gene_2	.	-
    <BLANKLINE>

    Use a 3kb merge distance so the last 2 features are merged together:

    >>> x = tsses(db, merge_overlapping=True, merge_kwargs=dict(d=3000))
    >>> print(x)  # doctest: +NORMALIZE_WHITESPACE
    chr2L	7528	7529	FBgn0031208	.	+
    chr2L	10999	12500	Fk_gene_1,Fk_gene_2	.	-
    <BLANKLINE>


    The set of unique TSSes for each gene, +1kb upstream and 500bp downstream:

    >>> x = tsses(db, merge_overlapping=True)
    >>> x = x.slop(l=1000, r=500, s=True, genome='dm3')
    >>> print(x)  # doctest: +NORMALIZE_WHITESPACE
    chr2L	6528	8029	FBgn0031208	.	+
    chr2L	10499	12000	Fk_gene_1	.	-
    chr2L	11999	13500	Fk_gene_2	.	-
    <BLANKLINE>


    """
    _override = os.environ.get('GFFUTILS_USES_BEDTOOLS_227_OR_LATER', None)
    if _override is not None:
        if _override == 'true':
            bedtools_227_or_later = True
        elif _override == 'false':
            bedtools_227_or_later = False
        else:
            raise ValueError(
                "Unknown value for GFFUTILS_USES_BEDTOOLS_227_OR_LATER "
                "environment variable: {0}".format(_override))

    if bedtools_227_or_later:
        _merge_kwargs = dict(o='distinct', s=True, c='4,5,6')
    else:
        _merge_kwargs = dict(o='distinct', s=True, c='4')

    if merge_kwargs is not None:
        _merge_kwargs.update(merge_kwargs)

    def gen():
        """
        Generator of pybedtools.Intervals representing TSSes.
        """
        for gene in db.features_of_type('gene'):
            for transcript in db.children(gene, level=1):
                if transcript.strand == '-':
                    transcript.start = transcript.stop
                else:
                    transcript.stop = transcript.start
                transcript.featuretype = transcript.featuretype + '_TSS'
                yield helpers.asinterval(transcript)

    # GFF/GTF format
    x = pybedtools.BedTool(gen()).sort()

    # Figure out default attrs to use, depending on the original format.
    if attrs is None:
        if db.dialect['fmt'] == 'gtf':
            attrs = 'gene_id'
        else:
            attrs = 'ID'

    if merge_overlapping or as_bed6:

        if isinstance(attrs, agouti_pkg.six.string_types):
            attrs = [attrs]

        def to_bed(f):
            """
            Given a pybedtools.Interval, return a new Interval with the name
            set according to the kwargs provided above.
            """
            name = attrs_sep.join([f.attrs[i] for i in attrs])
            return pybedtools.Interval(
                f.chrom,
                f.start,
                f.stop,
                name,
                str(f.score),
                f.strand)

        x = x.each(to_bed).saveas()

    if merge_overlapping:
        if bedtools_227_or_later:
            x = x.merge(**_merge_kwargs)
        else:
            def fix_merge(f):
                f = featurefuncs.extend_fields(f, 6)
                return pybedtools.Interval(
                    f.chrom,
                    f.start,
                    f.stop,
                    f[4],
                    '.',
                    f[3])
            x = x.merge(**_merge_kwargs).saveas().each(fix_merge).saveas()

    return x



import os
import agouti_pkg.six
import sqlite3
import shutil
import warnings
from agouti_pkg.gffutils import bins
from agouti_pkg.gffutils import helpers
from agouti_pkg.gffutils import constants
from agouti_pkg.gffutils.feature import Feature
from agouti_pkg.gffutils.exceptions import FeatureNotFoundError


class FeatureDB(object):
    # docstring to be filled in for methods that call out to
    # helpers.make_query()
    _method_doc = """
        limit : string or tuple
            Limit the results to a genomic region.  If string, then of the form
            "seqid:start-end"; if tuple, then (seqid, start, end).

        strand : "-" | "+" | "."
            Limit the results to one strand

        featuretype : string or tuple
            Limit the results to one or several featuretypes.

        order_by : string or tuple
            Order results by one or many fields; the string or tuple items must
            be in: 'seqid', 'source', 'featuretype', 'start', 'end', 'score',
            'strand', 'frame', 'attributes', 'extra'.

        reverse : bool
            Change sort order; only relevant if `order_by` is not None.  By
            default, results will be in ascending order, so use `reverse=True`
            for descending.

        completely_within : bool
            If False (default), a single bp overlap with `limit` is sufficient
            to return a feature; if True, then the feature must be completely
            within `limit`. Only relevant when `limit` is not None.
    """

    def __init__(self, dbfn, default_encoding='utf-8', keep_order=False,
                 pragmas=constants.default_pragmas,
                 sort_attribute_values=False,
                 text_factory=sqlite3.OptimizedUnicode):
        """
        Connect to a database created by :func:`gffutils.create_db`.

        Parameters
        ----------

        dbfn : str

            Path to a database created by :func:`gffutils.create_db`.

        text_factory : callable

            Optionally set the way sqlite3 handles strings.  Default is
            sqlite3.OptimizedUnicode, which returns ascii when possible,
            unicode otherwise

        encoding : str

            When non-ASCII characters are encountered, assume they are in this
            encoding.

        keep_order : bool

            If True, all features returned from this instance will have the
            order of their attributes maintained.  This can be turned on or off
            database-wide by setting the `keep_order` attribute or with this
            kwarg, or on a feature-by-feature basis by setting the `keep_order`
            attribute of an individual feature.

            Default is False, since this includes a sorting step that can get
            time-consuming for many features.

        pragmas : dict
            Dictionary of pragmas to use when connecting to the database.  See
            http://www.sqlite.org/pragma.html for the full list of
            possibilities, and constants.default_pragmas for the defaults.
            These can be changed later using the :meth:`FeatureDB.set_pragmas`
            method.

        Notes
        -----

            `dbfn` can also be a subclass of :class:`_DBCreator`, useful for
            when :func:`gffutils.create_db` is provided the ``dbfn=":memory:"``
            kwarg.

        """
        # Since specifying the string ":memory:" will actually try to connect
        # to a new, separate (and empty) db in memory, we can alternatively
        # pass in a sqlite connection instance to use its existing, in-memory
        # db.
        from agouti_pkg.gffutils import create
        if isinstance(dbfn, create._DBCreator):
            self.conn = dbfn.conn
            self.dbfn = dbfn.dbfn

        elif isinstance(dbfn, sqlite3.Connection):
            self.conn = dbfn
            self.dbfn = dbfn
        # otherwise assume dbfn is a string.
        elif dbfn == ':memory:':
            raise ValueError(
                "cannot connect to memory db; please provide the connection")
        else:
            if not os.path.exists(dbfn):
                raise ValueError("Database file %s does not exist" % dbfn)
            self.dbfn = dbfn
            self.conn = sqlite3.connect(self.dbfn)

        if text_factory is not None:
            self.conn.text_factory = text_factory
        self.conn.row_factory = sqlite3.Row

        self.default_encoding = default_encoding
        self.keep_order = keep_order
        self.sort_attribute_values = sort_attribute_values
        c = self.conn.cursor()

        # Load some meta info
        # TODO: this is a good place to check for previous versions, and offer
        # to upgrade...
        c.execute(
            '''
            SELECT version, dialect FROM meta
            ''')
        version, dialect = c.fetchone()
        self.version = version
        self.dialect = helpers._unjsonify(dialect)

        # Load directives from db
        c.execute(
            '''
            SELECT directive FROM directives
            ''')
        self.directives = [directive[0] for directive in c if directive]

        # Load autoincrements so that when we add new features, we can start
        # autoincrementing from where we last left off (instead of from 1,
        # which would cause name collisions)
        c.execute(
            '''
            SELECT base, n FROM autoincrements
            ''')
        self._autoincrements = dict(c)

        self.set_pragmas(pragmas)

        if not self._analyzed():
            warnings.warn(
                "It appears that this database has not had the ANALYZE "
                "sqlite3 command run on it. Doing so can dramatically "
                "speed up queries, and is done by default for databases "
                "created with gffutils >0.8.7.1 (this database was "
                "created with version %s) Consider calling the analyze() "
                "method of this object." % self.version)

    def set_pragmas(self, pragmas):
        """
        Set pragmas for the current database connection.

        Parameters
        ----------
        pragmas : dict
            Dictionary of pragmas; see constants.default_pragmas for a template
            and http://www.sqlite.org/pragma.html for a full list.
        """
        self.pragmas = pragmas
        c = self.conn.cursor()
        c.executescript(
            ';\n'.join(
                ['PRAGMA %s=%s' % i for i in self.pragmas.items()]
            )
        )
        self.conn.commit()

    def _feature_returner(self, **kwargs):
        """
        Returns a feature, adding additional database-specific defaults
        """
        kwargs.setdefault('dialect', self.dialect)
        kwargs.setdefault('keep_order', self.keep_order)
        kwargs.setdefault('sort_attribute_values', self.sort_attribute_values)
        return Feature(**kwargs)

    def _analyzed(self):
        res = self.execute(
            """
            SELECT name FROM sqlite_master WHERE type='table'
            AND name='sqlite_stat1';
            """)
        return len(list(res)) == 1

    def schema(self):
        """
        Returns the database schema as a string.
        """
        c = self.conn.cursor()
        c.execute(
            '''
            SELECT sql FROM sqlite_master
            ''')
        results = []
        for i, in c:
            if i is not None:
                results.append(i)
        return '\n'.join(results)

    def __getitem__(self, key):
        if isinstance(key, Feature):
            key = key.id
        c = self.conn.cursor()
        try:
            c.execute(constants._SELECT + ' WHERE id = ?', (key,))
        except sqlite3.ProgrammingError:
            c.execute(
                constants._SELECT + ' WHERE id = ?',
                (key.decode(self.default_encoding),))
        results = c.fetchone()
        # TODO: raise error if more than one key is found
        if results is None:
            raise FeatureNotFoundError(key)
        return self._feature_returner(**results)

    def count_features_of_type(self, featuretype=None):
        """
        Simple count of features.

        Can be faster than "grep", and is faster than checking the length of
        results from :meth:`gffutils.FeatureDB.features_of_type`.

        Parameters
        ----------

        featuretype : string

            Feature type (e.g., "gene") to count.  If None, then count *all*
            features in the database.

        Returns
        -------
        The number of features of this type, as an integer

        """
        c = self.conn.cursor()
        if featuretype is not None:
            c.execute(
                '''
                SELECT count() FROM features
                WHERE featuretype = ?
                ''', (featuretype,))
        else:
            c.execute(
                '''
                SELECT count() FROM features
                ''')

        results = c.fetchone()
        if results is not None:
            results = results[0]
        return results

    def features_of_type(self, featuretype, limit=None, strand=None,
                         order_by=None, reverse=False,
                         completely_within=False):
        """
        Returns an iterator of :class:`gffutils.Feature` objects.

        Parameters
        ----------
        {_method_doc}
        """
        query, args = helpers.make_query(
            args=[],
            limit=limit,
            featuretype=featuretype,
            order_by=order_by,
            reverse=reverse,
            strand=strand,
            completely_within=completely_within,
        )

        for i in self._execute(query, args):
            yield self._feature_returner(**i)

    # TODO: convert this to a syntax similar to itertools.groupby
    def iter_by_parent_childs(self, featuretype="gene", level=None,
                              order_by=None, reverse=False,
                              completely_within=False):
        """
        For each parent of type `featuretype`, yield a list L of that parent
        and all of its children (`[parent] + list(children)`). The parent will
        always be L[0].

        This is useful for "sanitizing" a GFF file for downstream tools.

        Additional kwargs are passed to :meth:`FeatureDB.children`, and will
        therefore only affect items L[1:] in each yielded list.
        """
        # Get all the parent records of the requested feature type
        parent_recs = self.all_features(featuretype=featuretype)
        # Return a generator of these parent records and their
        # children
        for parent_rec in parent_recs:
            unit_records = \
                [parent_rec] + list(self.children(parent_rec.id, level = level))
            yield unit_records

    def all_features(self, limit=None, strand=None, featuretype=None,
                     order_by=None, reverse=False, completely_within=False):
        """
        Iterate through the entire database.

        Returns
        -------
        A generator object that yields :class:`Feature` objects.

        Parameters
        ----------
        {_method_doc}
        """
        query, args = helpers.make_query(
            args=[],
            limit=limit,
            strand=strand,
            featuretype=featuretype,
            order_by=order_by,
            reverse=reverse,
            completely_within=completely_within
        )
        for i in self._execute(query, args):
            yield self._feature_returner(**i)

    def featuretypes(self):
        """
        Iterate over feature types found in the database.

        Returns
        -------
        A generator object that yields featuretypes (as strings)
        """
        c = self.conn.cursor()
        c.execute(
            '''
            SELECT DISTINCT featuretype from features
            ''')
        for i, in c:
            yield i

    def _relation(self, id, join_on, join_to, level=None, featuretype=None,
                  order_by=None, reverse=False, completely_within=False,
                  limit=None):

        # The following docstring will be included in the parents() and
        # children() docstrings to maintain consistency, since they both
        # delegate to this method.
        """
        Parameters
        ----------

        id : string or a Feature object

        level : None or int

            If `level=None` (default), then return all children regardless
            of level.  If `level` is an integer, then constrain to just that
            level.
        {_method_doc}

        Returns
        -------
        A generator object that yields :class:`Feature` objects.
        """

        if isinstance(id, Feature):
            id = id.id

        other = '''
        JOIN relations
        ON relations.{join_on} = features.id
        WHERE relations.{join_to} = ?
        '''.format(**locals())
        args = [id]

        level_clause = ''
        if level is not None:
            level_clause = 'relations.level = ?'
            args.append(level)

        query, args = helpers.make_query(
            args=args,
            other=other,
            extra=level_clause,
            featuretype=featuretype,
            order_by=order_by,
            reverse=reverse,
            limit=limit,
            completely_within=completely_within,
        )

        # modify _SELECT so that only unique results are returned
        query = query.replace("SELECT", "SELECT DISTINCT")
        for i in self._execute(query, args):
            yield self._feature_returner(**i)

    def children(self, id, level=None, featuretype=None, order_by=None,
                 reverse=False, limit=None, completely_within=False):
        """
        Return children of feature `id`.
        {_relation_docstring}
        """
        return self._relation(
            id, join_on='child', join_to='parent', level=level,
            featuretype=featuretype, order_by=order_by, reverse=reverse,
            limit=limit, completely_within=completely_within)

    def parents(self, id, level=None, featuretype=None, order_by=None,
                reverse=False, completely_within=False, limit=None):
        """
        Return parents of feature `id`.
        {_relation_docstring}
        """
        return self._relation(
            id, join_on='parent', join_to='child', level=level,
            featuretype=featuretype, order_by=order_by, reverse=reverse,
            limit=limit, completely_within=completely_within)

    def _execute(self, query, args):
        self._last_query = query
        self._last_args = args
        c = self.conn.cursor()
        c.execute(query, tuple(args))
        return c

    def execute(self, query):
        """
        Execute arbitrary queries on the db.

        .. seealso::

                :class:`FeatureDB.schema` may be helpful when writing your own
                queries.

        Parameters
        ----------

        query : str

            Query to execute -- trailing ";" optional.

        Returns
        -------
        A sqlite3.Cursor object that can be iterated over.
        """
        c = self.conn.cursor()
        return c.execute(query)

    def analyze(self):
        """
        Runs the sqlite ANALYZE command to potentially speed up queries
        dramatically.
        """
        self.execute('ANALYZE features')
        self.conn.commit()

    def region(self, region=None, seqid=None, start=None, end=None,
               strand=None, featuretype=None, completely_within=False):
        """
        Return features within specified genomic coordinates.

        Specifying genomic coordinates can be done in a flexible manner

        Parameters
        ----------
        region : string, tuple, or Feature instance
            If string, then of the form "seqid:start-end".  If tuple, then
            (seqid, start, end).  If :class:`Feature`, then use the features
            seqid, start, and end values.

            This argument is mutually exclusive with start/end/seqid.

            *Note*: By design, even if a feature is provided, its strand will
            be ignored.  If you want to restrict the output by strand, use the
            separate `strand` kwarg.

        strand : + | - | . | None
            If `strand` is provided, then only those features exactly matching
            `strand` will be returned. So `strand='.'` will only return
            unstranded features. Default is `strand=None` which does not
            restrict by strand.

        seqid, start, end, strand
            Mutually exclusive with `region`.  These kwargs can be used to
            approximate slice notation; see "Details" section below.

        featuretype : None, string, or iterable
            If not None, then restrict output.  If string, then only report
            that feature type.  If iterable, then report all featuretypes in
            the iterable.

        completely_within : bool
            By default (`completely_within=False`), returns features that
            partially or completely overlap `region`.  If
            `completely_within=True`, features that are completely within
            `region` will be returned.

        Notes
        -------

        The meaning of `seqid`, `start`, and `end` is interpreted as follows:

        ====== ====== ===== ======================================
        seqid  start  end   meaning
        ====== ====== ===== ======================================
        str    int    int   equivalent to `region` kwarg
        None   int    int   features from all chroms within coords
        str    None   int   equivalent to [:end] slice notation
        str    int    None  equivalent to [start:] slice notation
        None   None   None  equivalent to FeatureDB.all_features()
        ====== ====== ===== ======================================

        If performance is a concern, use `completely_within=True`. This allows
        the query to be optimized by only looking for features that fall in the
        precise genomic bin (same strategy as UCSC Genome Browser and
        BEDTools). Otherwise all features' start/stop coords need to be
        searched to see if they partially overlap the region of interest.

        Examples
        --------

        - `region(seqid="chr1", start=1000)` returns all features on chr1 that
          start or extend past position 1000

        - `region(seqid="chr1", start=1000, completely_within=True)` returns
          all features on chr1 that start past position 1000.

        - `region("chr1:1-100", strand="+", completely_within=True)` returns
          only plus-strand features that completely fall within positions 1 to
          100 on chr1.

        Returns
        -------
        A generator object that yields :class:`Feature` objects.
        """
        # Argument handling.
        if region is not None:
            if (seqid is not None) or (start is not None) or (end is not None):
                raise ValueError(
                    "If region is supplied, do not supply seqid, "
                    "start, or end as separate kwargs")
            if isinstance(region, agouti_pkg.six.string_types):
                toks = region.split(':')
                if len(toks) == 1:
                    seqid = toks[0]
                    start, end = None, None
                else:
                    seqid, coords = toks[:2]
                    if len(toks) == 3:
                        strand = toks[2]
                    start, end = coords.split('-')

            elif isinstance(region, Feature):
                seqid = region.seqid
                start = region.start
                end = region.end
                strand = region.strand

            # otherwise assume it's a tuple
            else:
                seqid, start, end = region[:3]

        # e.g.,
        #   completely_within=True..... start >= {start} AND end <= {end}
        #   completely_within=False.... start <  {end}   AND end >  {start}
        if completely_within:
            start_op = '>='
            end_op = '<='
        else:
            start_op = '<'
            end_op = '>'
            end, start = start, end

        args = []
        position_clause = []
        if seqid is not None:
            position_clause.append('seqid = ?')
            args.append(seqid)
        if start is not None:
            start = int(start)
            position_clause.append('start %s ?' % start_op)
            args.append(start)
        if end is not None:
            end = int(end)
            position_clause.append('end %s ?' % end_op)
            args.append(end)

        position_clause = ' AND '.join(position_clause)

        # Only use bins if we have defined boundaries and completely_within is
        # True. Otherwise you can't know how far away a feature stretches
        # (which means bins are not computable ahead of time)
        if (start is not None) and (end is not None) and completely_within:
            _bins = list(bins.bins(start, end, one=False))
            # See issue #45
            if len(_bins) < 900:
                _bin_clause = ' or ' .join(['bin = ?' for _ in _bins])
                _bin_clause = 'AND ( %s )' % _bin_clause
                args += _bins
            else:
                _bin_clause = ''
        else:
            _bin_clause = ''

        query = ' '.join([
            constants._SELECT,
            'WHERE ',
            position_clause,
            _bin_clause])

        # Add the featuretype clause
        if featuretype is not None:
            if isinstance(featuretype, agouti_pkg.six.string_types):
                featuretype = [featuretype]
            feature_clause = ' or '.join(
                ['featuretype = ?' for _ in featuretype])
            query += ' AND (%s) ' % feature_clause
            args.extend(featuretype)

        if strand is not None:
            strand_clause = ' and strand = ? '
            query += strand_clause
            args.append(strand)

        c = self.conn.cursor()
        self._last_query = query
        self._last_args = args
        self._context = {
            'start': start,
            'end': end,
            'seqid': seqid,
            'region': region,
        }
        c.execute(query, tuple(args))
        for i in c:
            yield self._feature_returner(**i)

    def interfeatures(self, features, new_featuretype=None,
                      merge_attributes=True, dialect=None,
                      attribute_func=None, update_attributes=None):
        """
        Construct new features representing the space between features.

        For example, if `features` is a list of exons, then this method will
        return the introns.  If `features` is a list of genes, then this method
        will return the intergenic regions.

        Providing N features will return N - 1 new features.

        This method purposefully does *not* do any merging or sorting of
        coordinates, so you may want to use :meth:`FeatureDB.merge` first.

        The new features' attributes will be a merge of the neighboring
        features' attributes.  This is useful if you have provided a list of
        exons; the introns will then retain the transcript and/or gene parents.

        Parameters
        ----------
        features : iterable of :class:`feature.Feature` instances
            Sorted, merged iterable

        new_featuretype : string or None
            The new features will all be of this type, or, if None (default)
            then the featuretypes will be constructed from the neighboring
            features, e.g., `inter_exon_exon`.

        attribute_func : callable or None
            If None, then nothing special is done to the attributes.  If
            callable, then the callable accepts two attribute dictionaries and
            returns a single attribute dictionary.  If `merge_attributes` is
            True, then `attribute_func` is called before `merge_attributes`.
            This could be useful for manually managing IDs for the new
            features.

        update_attributes : dict
            After attributes have been modified and merged, this dictionary can
            be used to replace parts of the attributes dictionary.

        Returns
        -------
        A generator that yields :class:`Feature` objects
        """
        for i, f in enumerate(features):
            # no inter-feature for the first one
            if i == 0:
                interfeature_start = f.stop
                last_feature = f
                continue

            interfeature_stop = f.start
            if new_featuretype is None:
                new_featuretype = 'inter_%s_%s' % (
                    last_feature.featuretype, f.featuretype)
            assert last_feature.strand == f.strand
            assert last_feature.chrom == f.chrom
            strand = last_feature.strand
            chrom = last_feature.chrom

            # Shrink
            interfeature_start += 1
            interfeature_stop -= 1

            if merge_attributes:
                new_attributes = helpers.merge_attributes(
                    last_feature.attributes, f.attributes)
            else:
                new_attributes = {}

            if update_attributes:
                new_attributes.update(update_attributes)

            new_bin = bins.bins(
                interfeature_start, interfeature_stop, one=True)
            _id = None
            fields = dict(
                seqid=chrom,
                source='gffutils_derived',
                featuretype=new_featuretype,
                start=interfeature_start,
                end=interfeature_stop,
                score='.',
                strand=strand,
                frame='.',
                attributes=new_attributes,
                bin=new_bin)

            if dialect is None:
                # Support for @classmethod -- if calling from the class, then
                # self.dialect is not defined, so defer to Feature's default
                # (which will be constants.dialect, or GFF3).
                try:
                    dialect = self.dialect
                except AttributeError:
                    dialect = None
            yield self._feature_returner(**fields)
            interfeature_start = f.stop

    def delete(self, features, make_backup=True, **kwargs):
        """
        Delete features from database.

        features : str, iterable, FeatureDB instance
            If FeatureDB, all features will be used. If string, assume it's the
            ID of the feature to remove. Otherwise, assume it's an iterable of
            Feature objects. The classes in gffutils.iterators may be helpful
            in this case.

        make_backup : bool
            If True, and the database you're about to update is a file on disk,
            makes a copy of the existing database and saves it with a .bak
            extension.

        Returns
        -------
        FeatureDB object, with features deleted.
        """
        if make_backup:
            if isinstance(self.dbfn, agouti_pkg.six.string_types):
                shutil.copy2(self.dbfn, self.dbfn + '.bak')

        c = self.conn.cursor()
        query1 = """
        DELETE FROM features WHERE id = ?
        """
        query2 = """
        DELETE FROM relations WHERE parent = ? OR child = ?
        """
        if isinstance(features, FeatureDB):
            features = features.all_features()
        if isinstance(features, agouti_pkg.six.string_types):
            features = [features]
        if isinstance(features, Feature):
            features = [features]
        for feature in features:
            if isinstance(feature, agouti_pkg.six.string_types):
                _id = feature
            else:
                _id = feature.id
            c.execute(query1, (_id,))
            c.execute(query2, (_id, _id))
        self.conn.commit()
        return self

    def update(self, data, make_backup=True, **kwargs):
        """
        Update database with features in `data`.

        data : str, iterable, FeatureDB instance
            If FeatureDB, all data will be used. If string, assume it's
            a filename of a GFF or GTF file.  Otherwise, assume it's an
            iterable of Feature objects.  The classes in gffutils.iterators may
            be helpful in this case.

        make_backup : bool
            If True, and the database you're about to update is a file on disk,
            makes a copy of the existing database and saves it with a .bak
            extension.

        Notes
        -----
        Other kwargs are used in the same way as in gffutils.create_db; see the
        help for that function for details.

        Returns
        -------
        FeatureDB with updated features.
        """
        from gffutils import create
        from gffutils import iterators
        if make_backup:
            if isinstance(self.dbfn, agouti_pkg.six.string_types):
                shutil.copy2(self.dbfn, self.dbfn + '.bak')

        # get iterator-specific kwargs
        _iterator_kwargs = {}
        for k, v in kwargs.items():
            if k in constants._iterator_kwargs:
                _iterator_kwargs[k] = v

        # Handle all sorts of input
        data = iterators.DataIterator(data, **_iterator_kwargs)

        if self.dialect['fmt'] == 'gtf':
            if 'id_spec' not in kwargs:
                kwargs['id_spec'] = {
                    'gene': 'gene_id', 'transcript': 'transcript_id'}
            db = create._GTFDBCreator(
                data=data, dbfn=self.dbfn, dialect=self.dialect, **kwargs)
        elif self.dialect['fmt'] == 'gff3':
            if 'id_spec' not in kwargs:
                kwargs['id_spec'] = 'ID'
            db = create._GFFDBCreator(
                data=data, dbfn=self.dbfn, dialect=self.dialect, **kwargs)

        else:
            raise ValueError

        db._populate_from_lines(data)
        db._update_relations()
        db._finalize()
        return db

    def add_relation(self, parent, child, level, parent_func=None,
                     child_func=None):
        """
        Manually add relations to the database.

        Parameters
        ----------
        parent : str or Feature instance
             Parent feature to add.

        child : str or Feature instance
            Child feature to add

        level : int
            Level of the relation.  For example, if parent is a gene and child
            is an mRNA, then you might want level to be 1.  But if child is an
            exon, then level would be 2.

        parent_func, child_func : callable
            These optional functions control how attributes are updated in the
            database.  They both have the signature `func(parent, child)` and
            must return a [possibly modified] Feature instance.  For example,
            we could add the child's database id as the "child" attribute in
            the parent::

                def parent_func(parent, child):
                    parent.attributes['child'] = child.id

            and add the parent's "gene_id" as the child's "Parent" attribute::

                def child_func(parent, child):
                    child.attributes['Parent'] = parent['gene_id']

        Returns
        -------
        FeatureDB object with new relations added.
        """
        if isinstance(parent, agouti_pkg.six.string_types):
            parent = self[parent]
        if isinstance(child, agouti_pkg.six.string_types):
            child = self[child]

        c = self.conn.cursor()
        c.execute('''
                  INSERT INTO relations (parent, child, level)
                  VALUES (?, ?, ?)''', (parent.id, child.id, level))

        if parent_func is not None:
            parent = parent_func(parent, child)
            self._update(parent, c)
        if child_func is not None:
            child = child_func(parent, child)
            self._update(child, c)

        self.conn.commit()
        return self

    def _update(self, feature, cursor):
        values = [list(feature.astuple()) + [feature.id]]
        cursor.execute(
            constants._UPDATE, *tuple(values))

    def _insert(self, feature, cursor):
        """
        Insert a feature into the database.
        """
        try:
            cursor.execute(constants._INSERT, feature.astuple())
        except sqlite3.ProgrammingError:
            cursor.execute(
                constants._INSERT, feature.astuple(self.default_encoding))

    def create_introns(self, exon_featuretype='exon',
                       grandparent_featuretype='gene', parent_featuretype=None,
                       new_featuretype='intron', merge_attributes=True):
        """
        Create introns from existing annotations.


        Parameters
        ----------
        exon_featuretype : string
            Feature type to use in order to infer introns.  Typically `"exon"`.

        grandparent_featuretype : string
            If `grandparent_featuretype` is not None, then group exons by
            children of this featuretype.  If `granparent_featuretype` is
            "gene" (default), then introns will be created for all first-level
            children of genes.  This may include mRNA, rRNA, ncRNA, etc.  If
            you only want to infer introns from one of these featuretypes
            (e.g., mRNA), then use the `parent_featuretype` kwarg which is
            mutually exclusive with `grandparent_featuretype`.

        parent_featuretype : string
            If `parent_featuretype` is not None, then only use this featuretype
            to infer introns.  Use this if you only want a subset of
            featuretypes to have introns (e.g., "mRNA" only, and not ncRNA or
            rRNA). Mutually exclusive with `grandparent_featuretype`.

        new_featuretype : string
            Feature type to use for the inferred introns; default is
            `"intron"`.

        merge_attributes : bool
            Whether or not to merge attributes from all exons. If False then no
            attributes will be created for the introns.

        Returns
        -------
        A generator object that yields :class:`Feature` objects representing
        new introns

        Notes
        -----
        The returned generator can be passed directly to the
        :meth:`FeatureDB.update` method to permanently add them to the
        database, e.g., ::

            db.update(db.create_introns())

        """
        if (grandparent_featuretype and parent_featuretype) or (
            grandparent_featuretype is None and parent_featuretype is None
        ):
            raise ValueError("exactly one of `grandparent_featuretype` or "
                             "`parent_featuretype` should be provided")

        if grandparent_featuretype:
            def child_gen():
                for gene in self.features_of_type(grandparent_featuretype):
                    for child in self.children(gene, level=1):
                        yield child
        elif parent_featuretype:
            def child_gen():
                for child in self.features_of_type(parent_featuretype):
                    yield child

        for child in child_gen():
            exons = self.children(child, level=1, featuretype=exon_featuretype,
                                  order_by='start')
            for intron in self.interfeatures(
                exons, new_featuretype=new_featuretype,
                merge_attributes=merge_attributes, dialect=self.dialect
            ):
                yield intron

    def merge(self, features, ignore_strand=False):
        """
        Merge overlapping features together.

        Parameters
        ----------

        features : iterator of Feature instances

        ignore_strand : bool
            If True, features on multiple strands will be merged, and the final
            strand will be set to '.'.  Otherwise, ValueError will be raised if
            trying to merge features on differnt strands.

        Returns
        -------
        A generator object that yields :class:`Feature` objects representing
        the newly merged features.
        """

        # Consume iterator up front...
        features = list(features)

        if len(features) == 0:
            raise StopIteration

        # Either set all strands to '+' or check for strand-consistency.
        if ignore_strand:
            strand = '.'
        else:
            strands = [i.strand for i in features]
            if len(set(strands)) > 1:
                raise ValueError('Specify ignore_strand=True to force merging '
                                 'of multiple strands')
            strand = strands[0]

        # Sanity check to make sure all features are from the same chromosome.
        chroms = [i.chrom for i in features]
        if len(set(chroms)) > 1:
            raise NotImplementedError('Merging multiple chromosomes not '
                                      'implemented')
        chrom = chroms[0]

        # To start, we create a merged feature of just the first feature.
        current_merged_start = features[0].start
        current_merged_stop = features[0].stop

        # We don't need to check the first one, so start at feature #2.
        for feature in features[1:]:
            # Does this feature start within the currently merged feature?...
            if feature.start <= current_merged_stop + 1:
                # ...It starts within, so leave current_merged_start where it
                # is.  Does it extend any farther?
                if feature.stop >= current_merged_stop:
                    # Extends further, so set a new stop position
                    current_merged_stop = feature.stop
                else:
                    # If feature.stop < current_merged_stop, it's completely
                    # within the previous feature.  Nothing more to do.
                    continue
            else:
                # The start position is outside the merged feature, so we're
                # done with the current merged feature.  Prepare for output...
                merged_feature = dict(
                    seqid=feature.chrom,
                    source='.',
                    featuretype=feature.featuretype,
                    start=current_merged_start,
                    end=current_merged_stop,
                    score='.',
                    strand=strand,
                    frame='.',
                    attributes='')
                yield self._feature_returner(**merged_feature)

                # and we start a new one, initializing with this feature's
                # start and stop.
                current_merged_start = feature.start
                current_merged_stop = feature.stop

        # need to yield the last one.
        if len(features) == 1:
            feature = features[0]
        merged_feature = dict(
            seqid=feature.chrom,
            source='.',
            featuretype=feature.featuretype,
            start=current_merged_start,
            end=current_merged_stop,
            score='.',
            strand=strand,
            frame='.',
            attributes='')
        yield self._feature_returner(**merged_feature)

    def children_bp(self, feature, child_featuretype='exon', merge=False,
                    ignore_strand=False):
        """
        Total bp of all children of a featuretype.

        Useful for getting the exonic bp of an mRNA.

        Parameters
        ----------

        feature : str or Feature instance

        child_featuretype : str
            Which featuretype to consider.  For example, to get exonic bp of an
            mRNA, use `child_featuretype='exon'`.

        merge : bool
            Whether or not to merge child features together before summing
            them.

        ignore_strand : bool
            If True, then overlapping features on different strands will be
            merged together; otherwise, merging features with different strands
            will result in a ValueError.

        Returns
        -------
        Integer representing the total number of bp.
        """

        children = self.children(feature, featuretype=child_featuretype,
                                 order_by='start')
        if merge:
            children = self.merge(children, ignore_strand=ignore_strand)

        total = 0
        for child in children:
            total += len(child)
        return total

    def bed12(self, feature, block_featuretype=['exon'],
              thick_featuretype=['CDS'], thin_featuretype=None,
              name_field='ID', color=None):
        """
        Converts `feature` into a BED12 format.

        GFF and GTF files do not necessarily define genes consistently, so this
        method provides flexiblity in specifying what to call a "transcript".

        Parameters
        ----------
        feature : str or Feature instance
            In most cases, this feature should be a transcript rather than
            a gene.

        block_featuretype : str or list
            Which featuretype to use as the exons. These are represented as
            blocks in the BED12 format.  Typically 'exon'.

            Use the `thick_featuretype` and `thin_featuretype` arguments to
            control the display of CDS as thicker blocks and UTRs as thinner
            blocks.

            Note that the features for `thick` or `thin` are *not*
            automatically included in the blocks; if you do want them included,
            then those featuretypes should be added to this `block_features`
            list.

            If no child features of type `block_featuretype` are found, then
            the full `feature` is returned in BED12 format as if it had
            a single exon.

        thick_featuretype : str or list
            Child featuretype(s) to use in order to determine the boundaries of
            the "thick" blocks. In BED12 format, these represent coding
            sequences; typically this would be set to "CDS".  This argument is
            mutually exclusive with `thin_featuretype`.

            Specifically, the BED12 thickStart will be the start coord of the
            first `thick` item and the thickEnd will be the stop coord of the
            last `thick` item.

        thin_featuretype : str or list
            Child featuretype(s) to use in order to determine the boundaries of
            the "thin" blocks.  In BED12 format, these represent untranslated
            regions.  Typically "utr" or ['three_prime_UTR', 'five_prime_UTR'].
            Mutually exclusive with `thick_featuretype`.

            Specifically, the BED12 thickStart field will be the stop coord of
            the first `thin` item and the thickEnd field will be the start
            coord of the last `thin` item.

        name_field : str
            Which attribute of `feature` to use as the feature's name.  If this
            field is not present, a "." placeholder will be used instead.

        color : None or str
            If None, then use black (0,0,0) as the RGB color; otherwise this
            should be a comma-separated string of R,G,B values each of which
            are integers in the range 0-255.
        """
        if thick_featuretype and thin_featuretype:
            raise ValueError("Can only specify one of `thick_featuertype` or "
                             "`thin_featuretype`")

        exons = list(self.children(feature, featuretype=block_featuretype,
                                   order_by='start'))
        if len(exons) == 0:
            exons = [feature]
        feature = self[feature]
        first = exons[0].start
        last = exons[-1].stop

        if first != feature.start:
            raise ValueError(
                "Start of first exon (%s) does not match start of feature (%s)"
                % (first, feature.start))
        if last != feature.stop:
            raise ValueError(
                "End of last exon (%s) does not match end of feature (%s)"
                % (last, feature.stop))

        if color is None:
            color = '0,0,0'
        color = color.replace(' ', '').strip()

        # Use field names as defined at
        # http://genome.ucsc.edu/FAQ/FAQformat.html#format1
        chrom = feature.chrom
        chromStart = feature.start - 1
        chromEnd = feature.stop
        orig = constants.always_return_list
        constants.always_return_list = True
        try:
            name = feature[name_field][0]
        except KeyError:
            name = "."

        constants.always_return_list = orig

        score = feature.score
        if score == '.':
            score = '0'
        strand = feature.strand
        itemRgb = color
        blockCount = len(exons)
        blockSizes = [len(i) for i in exons]
        blockStarts = [i.start - 1 - chromStart for i in exons]

        if thick_featuretype:
            thick = list(self.children(feature, featuretype=thick_featuretype,
                                       order_by='start'))
            if len(thick) == 0:
                thickStart = feature.start
                thickEnd = feature.stop
            else:
                thickStart = thick[0].start - 1  # BED 0-based coords
                thickEnd = thick[-1].stop

        if thin_featuretype:
            thin = list(self.children(feature, featuretype=thin_featuretype,
                                      order_by='start'))
            if len(thin) == 0:
                thickStart = feature.start
                thickEnd = feature.stop
            else:
                thickStart = thin[0].stop
                thickEnd = thin[-1].start - 1  # BED 0-based coords

        tst = chromStart + blockStarts[-1] + blockSizes[-1]
        assert tst == chromEnd, "tst=%s; chromEnd=%s" % (tst, chromEnd)

        fields = [
            chrom,
            chromStart,
            chromEnd,
            name,
            score,
            strand,
            thickStart,
            thickEnd,
            itemRgb,
            blockCount,
            ','.join(map(str, blockSizes)),
            ','.join(map(str, blockStarts))]
        return '\t'.join(map(str, fields))

    # Recycle the docs for _relation so they stay consistent between parents()
    # and children()
    children.__doc__ = children.__doc__.format(
        _relation_docstring=_relation.__doc__)
    parents.__doc__ = parents.__doc__.format(
        _relation_docstring=_relation.__doc__)

    # Add the docs for methods that call helpers.make_query()
    for method in [parents, children, features_of_type, all_features]:
        method.__doc__ = method.__doc__.format(
            _method_doc=_method_doc)



"""
Test cases used for the nosetest generator over in parser_test.py.

Each item is a 3-tuple.  The first two items are the attribute string and the
expected parsed dictionary.  The third, if not None, is a reconstructed string
that is acceptable -- this is needed for cases like this:

    name "fgenesh1_pg.C_chr_1000007"; transcriptId 873

where there is not consistency in the quoting of values.  So, in this case, the
following would be an acceptable recontructed string (with quotes around the
873):

    name "fgenesh1_pg.C_chr_1000007"; transcriptId "873"


"""
attrs = [

    # c_elegans_WS199_shortened_gff.txt
    (
        "count=1;gene=amx-2;sequence=SAGE:ggcagagtcttttggca;"
        "transcript=B0019.1",

        {'count': ["1"],
         'gene': ['amx-2'],
         'sequence': ['SAGE:ggcagagtcttttggca'],
         'transcript': ['B0019.1'],
         },

        None,
    ),

    # ensembl_gtf.txt
    (
        'gene_id "Y74C9A.6"; transcript_id "Y74C9A.6"; exon_number "1"; '
        'gene_name "Y74C9A.6"; transcript_name "NR_001477.2";',

        {'gene_id': ['Y74C9A.6'],
         'transcript_id': ['Y74C9A.6'],
         'exon_number': ['1'],
         'gene_name': ['Y74C9A.6'],
         'transcript_name': ['NR_001477.2']},

        None,
    ),

    # F3-unique-3.v2.gff
    (
        'g=A3233312322232122211;i=1;p=1.000;q=23,12,18,17,10,24,19,14,27,9,23'
        ',9,16,20,11,7,8,4,4,14;u=0,0,0,1',

        {'g': ['A3233312322232122211'],
         'i': ['1'],
         'p': ['1.000'],
         'q': ['23',
               '12', '18', '17', '10', '24', '19', '14', '27', '9', '23', '9',
               '16', '20', '11', '7', '8', '4', '4', '14'],
         'u': ['0', '0', '0', '1']},

        None,
    ),


    # glimmer_nokeyval.gff3
    (
        "ID=GL0000006;Name=GL0000006;Lack 3'-end;",

        {'ID': ["GL0000006"],
         'Name': ['GL0000006'],
         "Lack 3'-end": []},

        None,
    ),

    # hybrid1.gff3
    (
        'ID=A00469;Dbxref=AFFX-U133:205840_x_at,Locuslink:2688,Genbank-mRNA:'
        'A00469,Swissprot:P01241,PFAM:PF00103,AFFX-U95:1332_f_at,Swissprot:'
        'SOMA_HUMAN;Note=growth%20hormone%201;Alias=GH1',

        {'ID': ['A00469'],
         'Dbxref': [
             'AFFX-U133:205840_x_at',
             'Locuslink:2688',
             'Genbank-mRNA:A00469',
             'Swissprot:P01241',
             'PFAM:PF00103',
             'AFFX-U95:1332_f_at',
             'Swissprot:SOMA_HUMAN',
         ],
         'Note': ['growth hormone 1'],
         'Alias': ['GH1']},

        'ID=A00469;Dbxref=AFFX-U133:205840_x_at,Locuslink:2688,Genbank-mRNA:'
        'A00469,Swissprot:P01241,PFAM:PF00103,AFFX-U95:1332_f_at,Swissprot:'
        'SOMA_HUMAN;Note=growth hormone 1;Alias=GH1',
    ),

    # jgi_gff2.txt
    #
    # This file is inconsitent with how it quotes values -- integers are not
    # quoted but string values are.  Only way to make this be invariant is to
    # keep track of the "flavor" of each attribute; not sure it's worth the
    # effort / processing time.
    (
        'name "fgenesh1_pg.C_chr_1000007"; transcriptId 873',

        {'name': ['fgenesh1_pg.C_chr_1000007'],
         'transcriptId': ['873']},

        'name "fgenesh1_pg.C_chr_1000007"; transcriptId "873"',
    ),

    # mouse_extra_comma.gff3: extra comma line
    #
    # Note extra empty string in the dictionary's "Parent" field.
    #
    (
        'Name=CDS:NC_000083.5:LOC100040603;Parent=XM_001475631.1,',

        {'Name': ['CDS:NC_000083.5:LOC100040603'],
         'Parent': ['XM_001475631.1', '']},

        None,
    ),

    # mouse_extra_comma.gff3
    #
    # Note the empty ID field.  Compare with the "Lack 3'-end" attribute of
    # glimmer_nokeyval.gff3 above.  Presumably the "Lack 3'-end" field should
    # be interpreted as "True", but an empty ID should be interpreted as "None"
    # or something.
    #
    # Furthermore, the "Lack 3'-end" has no trailing "=", but the "ID" field
    # here does.
    #
    # In both cases, the dictionary entry is simply an empty list; it's just in
    # the reconstruction where things get tricky.
    (
        'ID=;Parent=XM_001475631.1',

        {'ID': [],
         'Parent': ['XM_001475631.1']},

        'ID;Parent=XM_001475631.1',
    ),

    # ncbi_gff3.txt
    (
        'ID=NC_008596.1:speB:unknown_transcript_1;Parent=NC_008596.1:speB;'
        'locus_tag=MSMEG_1072;EC_number=3.5.3.11;note=identified%20by%20mat'
        'ch%20to%20protein%20family%20HMM%20PF00491%3B%20match%20to%20prote'
        'in%20family%20HMM%20TIGR01230;transl_table=11;product=agmatinase;p'
        'rotein_id=YP_885468.1;db_xref=GI:118469242;db_xref=GeneID:4535378;'
        'exon_number=1',

        {'ID': ['NC_008596.1:speB:unknown_transcript_1'],
         'Parent': ['NC_008596.1:speB'],
         'locus_tag': ['MSMEG_1072'],
         'EC_number': ['3.5.3.11'],
         'note': ['identified by match to protein family HMM P'
                  'F00491; match to protein family HMM TIGR01'
                  '230'],
         'transl_table': ['11'],
         'product': ['agmatinase'],
         'protein_id': ['YP_885468.1'],
         'db_xref': ['GI:118469242', 'GeneID:4535378'],
         'exon_number': ['1'],
         },

        'ID=NC_008596.1:speB:unknown_transcript_1;Parent=NC_008596.1:speB;'
        'locus_tag=MSMEG_1072;EC_number=3.5.3.11;note=identified by mat'
        'ch to protein family HMM PF00491%3B match to prote'
        'in family HMM TIGR01230;transl_table=11;product=agmatinase;p'
        'rotein_id=YP_885468.1;db_xref=GI:118469242;db_xref=GeneID:4535378;'
        'exon_number=1',
    ),

    # wormbase_gff2_alt.txt
    #
    (
        'CDS "cr01.sctg102.wum.2.1"',

        {'CDS': ['cr01.sctg102.wum.2.1']},

        None,
    ),

]



from agouti_pkg.miscallaneous import *


def prepare_output(lengths_dict, header, args, attributes_and_features,
                   database, processing_product, overlapping_features,
                   region, cds_start=None):
    """Prepares a single line of the output file. Parse ProcessingProduct\
        objects.

    Arguments:
        lengths_dict {dict} -- 5'UTR length, CDS length, 3'UTR length
        header {str} -- header
        args {argparse.Namespace} -- argparse command-line arguments
        attributes_and_features {dict} -- dictionary of attributes and\
            features that need to be annotated
        database {Database} -- Database
        processing_product {ProcessingProduct} -- object of ProcessingProduct
            class
        overlapping_features {list} -- list of overlapping features
        region {tuple} -- coordinates

    Keyword Arguments:
        cds_start {int} -- start of cds coordinate (default: {None})

    Returns:
        str -- single output line
    """

    out = ""

    if len(overlapping_features):  # if there are overlapping features

        for feature in overlapping_features:
            temp_out = "."
            temp_result = ""
            temp_featuretype="."
            temp_seqid="."
            temp_overlapping_feature_start="."
            temp_overlapping_feature_end="."

            if ((feature.strand != processing_product.strand) and
               args.strand_specific and not args.transcriptomic) or (args.transcriptomic and processing_product.strand == "-"):
                continue

            if (feature.featuretype in attributes_and_features.keys()):

                if not args.transcriptomic:

                    if ((args.completly_within and not completly_within(
                            region[1], region[2], feature)) or not
                            infer_feature_level(feature, database) ==
                            args.level):
                        continue

                    if args.level == 2:
                        gene_id = "{}\t".format((get_level1_parent(database,
                                                                   feature,
                                                                   id=True)))
                        out += "{bed}\t{featureid}\t{gene_id}\t{featuretype}\t\
                            {overlapping_feature_start}\t\
                            {overlapping_feature_end}\
                            ".format(bed=processing_product.bed_line.strip(),
                                     featureid=feature.id.strip(),
                                     gene_id=gene_id.strip(),
                                     featuretype=feature.featuretype.strip(),
                                     overlapping_feature_start=feature.start,
                                     overlapping_feature_end=feature.end)
                    else:
                        gene_id = ""
                        out += "{bed}\t{featureid}\t{featuretype}\t\
                                {overlapping_feature_start}\t\
                                {overlapping_feature_end}\
                                ".format(bed=processing_product.bed_line.strip(),
                                        featureid=feature.id.strip(),
                                        featuretype=feature.featuretype.strip(),
                                        overlapping_feature_start=feature.start,
                                        overlapping_feature_end=feature.end)

                    
                    out += "\t{}\n".format(
                        filter_overlapping_features(attributes_and_features,
                                                    args, database, feature,
                                                    overlapping_features,
                                                    header[1], header[2]))


                else:
                    comp_pos = completly_within_positive_coordinates  # abbr
                    temp = comp_pos(lengths_dict, feature, processing_product,
                                    cds_start)

                    if (not infer_feature_level(feature, database) ==
                        args.level or (args.completly_within and
                        processing_product.coords_outside_transcript ==
                        "just_one") or processing_product.coords_outside_transcript ==
                        "both") or (args.completly_within and temp[0] ==
                        "just_one") or temp == "both":
                        if args.level == 2:
                            temp_out = get_level1_parent(database, feature, id=True)
                            temp_featuretype=feature.featuretype
                            temp_seqid=feature.seqid
                            temp_overlapping_feature_start=feature.start
                            temp_overlapping_feature_end=feature.end
                            analyzed_set = set(attributes_and_features[feature.featuretype]).intersection(feature.attributes)
                            d = {}
                            for a in header[2]:
                                for attr in analyzed_set:
                                    if attr == a:
                                        d[attr] = ("").join(feature[attr])
                                # concatenation
                                temp_result = ""

                            for h in (header[1] + header[2]):
                                try:
                                    temp_result += "{}\t".format(d[h].strip())
                                except KeyError:
                                    temp_result += "{}\t".format(".")
                        continue

                    if args.level == 2:
                        gene_id = get_level1_parent(database, feature, id=True)
                    else:
                        gene_id = ""

                    out += "{bed}\t{gene_id}\t{featuretype}\t{seqid}\t\
                            {overlapping_feature_start}\t\
                            {overlapping_feature_end}\
                            ".format(bed=processing_product.bed_line.strip(),
                                     featuretype=feature.featuretype.strip(),
                                     gene_id=gene_id, seqid=feature.seqid.strip(),
                                     overlapping_feature_start=feature.start,
                                     overlapping_feature_end=feature.end)

                    filt_ov = filter_overlapping_features  # abbr
                    out += "\t{}\n".format(filt_ov(attributes_and_features,
                                                   args, database, feature, [],
                                                   header[1], header[2],
                                                   processing_product, region))

                    insert = processing_product.check_overlapping_feature__position(lengths_dict,
                                                                                    feature,
                                                                                    args.transcriptomic,
                                                                                    args.offset)

                    if (args.annotate_relative_location):
                        out = out.rstrip()
                        out += "\t{}\n".format(insert)

    if (out == ""):

        additional_fields = ""

        header_lengths = (len(header[1]) + len(header[2])) if not (
            args.annotate_relative_location) else (len(header[1]) +
                                                len(header[2]) + 1)

        if (args.transcriptomic):

            feature = database.database[processing_product.coordinates[0]]
            out += "{bed}\t{gene_id}\t{featuretype}\t{seqid}\t\
                    {overlapping_feature_start}\t{overlapping_feature_end}\
                    ".format(bed=processing_product.bed_line.strip(),
                            featuretype=temp_featuretype, gene_id=temp_out, seqid=temp_seqid,
                            overlapping_feature_start=temp_overlapping_feature_start,
                            overlapping_feature_end=temp_overlapping_feature_end)

            
            if args.level != 2 or ((feature.strand != processing_product.strand) and
               args.strand_specific and not args.transcriptomic) or (args.transcriptomic and processing_product.strand == "-"):
                out += "\t{}".format(".\t.\t.")
                for _ in range(0, len(header[2])):
                    out += "\t."

                out += "\n"
            else:
                out += "\t{}\n".format(temp_result)

            if (args.annotate_relative_location):
                out = out.rstrip()
                if not temp_featuretype == ".":
                    out += "\t{}\n".format(
                    processing_product.check_overlapping_feature__position(lengths_dict, feature,
                                                                           args.transcriptomic, args.offset))

        else:

            for _ in range(1, header_lengths):
                additional_fields += "\t."

            closest_gene = find_closest_gene(database, processing_product)
            find_closest_gene(database, processing_product)

            out += "{bed}\t{featureid}\t{gene_id}\t{featuretype}\t\
                    {overlapping_feature_start}\t{overlapping_feature_end}\
                    {add}".format(bed=processing_product.bed_line.strip(),
                                  featureid=closest_gene, featuretype=".",
                                  gene_id="intergenic",
                                  overlapping_feature_start=".",
                                  overlapping_feature_end=".",
                                  add=additional_fields)

    return out.strip()





from agouti_pkg.read_input import read_header_line


def prepare_bed_header(bed_file, custom, transcriptomic, num_of_bed_fields, sep, header_line_num):
    """Prepares part of the header for the features from BED line

    Arguments:
        bed_file {str} -- name or path of the input file
        custom {str} -- custom file format description - as in --custom arg
        transcriptomic {bool} -- args.transcriptomic
        num_of_bed_fields {int} -- number of columns in the input file
                sep {str} -- separator

    Returns:
        [str] -- bed header of the output file
    """

    rhl = read_header_line(bed_file, custom, sep, header_line_num)
    bed_header = ""

    if custom != "BED":
        idx = 1
        custom_format = custom.strip().split(",")
        custom_format = list(map(int, custom_format))

        if not transcriptomic:
            bed_fields = ["feature_id", "chr", "feature_start", "feature_end",
                          "strand"]
        else:
            bed_fields = ["feature_id", "transcript_id", "feature_start",
                          "feature_end", "strand"]

        indices = [index for index, value in sorted(enumerate(custom_format),
                   key=lambda x: x[1])]
        temp = 0

        for i in range(1, num_of_bed_fields + 1):

            if i not in custom_format:

                if not rhl[0]:  # if there is no header
                    bed_header = "{}\tCUSTOM_field_{}".format(bed_header, idx)
                else:
                    bed_header = "{}\t{}".format(bed_header, rhl[1][idx - 1])
                idx = idx + 1

            else:
                bed_header = "{}\t{}".format(bed_header,
                                             bed_fields[(indices[temp])])
                temp = temp + 1
    else:
        bed_fields = []

        if not transcriptomic:
            bed_fields = ["chr", "feature_start", "feature_end",
                          "feature_name", "feature_score", "strand"]
        else:
            bed_fields = ["transcript_id", "feature_start", "feature_end",
                          "feature_name", "feature_score", "strand"]

        for field in range(0, num_of_bed_fields):

            if field < 6:
                bed_header = "{}\t{}".format(bed_header, bed_fields[field])
            else:
                idx = field - 6

                if not rhl[0]:  # if there is no header
                    bed_header = "{}\tBED_field_{}".format(bed_header, idx + 1)
                else:
                    bed_header = "{}\t{}".format(bed_header, rhl[1][idx])

    return bed_header


def prepare_header(db, attributes_and_features, args, num_of_bed_fields, header_line_num):
    """Prepare header for the output file

    Arguments:
        db {Database} -- object of the Database class
        attributes_and_features {dict} -- dictionary of attributes and\
            features that need to be annotated
        args {argparse} -- parsed command line argument using argparse
        num_of_bed_fields {int} -- number of columns in the input file

    Returns:
        tuple -- header of the output file, attributes choosen by the user to\
            be annotated and present in the database,\
            attributes choosen by the user
    """

    features_to_annotate = set(db.features_at_3_level)
    features_choosen_by_user = set(attributes_and_features.keys())
    intersection = features_to_annotate & features_choosen_by_user

    if args.level == 1:
        features_at_given_level = db.features_at_1_level
    else:
        features_at_given_level = db.features_at_2_level

    attributes_choosen_by_user = set()

    for f in attributes_and_features.keys():
        if (f in features_at_given_level):
            for a in attributes_and_features[f]:
                attributes_choosen_by_user.add(a)

    attributes_choosen_by_user = sorted(attributes_choosen_by_user)
    intersection = list(intersection)
    intersection.sort()
    bed_header = prepare_bed_header(args.bed, args.custom, args.transcriptomic,
                                    num_of_bed_fields, args.sep, header_line_num)

    if (args.level == 1):
        header = "{}\tannotated_gene_id\tannotated_featuretype\
                  \tannotated_gene_start\tannotated_gene_end\
                  \t{}\t{}".format(bed_header, "\t".join(intersection),
                                   "\t".join(attributes_choosen_by_user))
    elif (args.level == 2):

        if args.transcriptomic:
            header = "{}\tannotated_gene_id\tannotated_featuretype\
                      \tannotated_chromosome\tannotated_transcript_start\
                      \tannotated_transcript_end\t{}\t{}\
                      ".format(bed_header, "\t".join(intersection),
                               "\t".join(attributes_choosen_by_user))
        else:
            header = "{}\tannotated_transcript_id\tannotated_gene_id\
                      \tannotated_featuretype\tannotated_transcript_start\
                      \tannotated_transcript_end\t{}\t{}\
                      ".format(bed_header, "\t".join(intersection),
                               "\t".join(attributes_choosen_by_user))

    if (args.annotate_relative_location):
        header = header.strip() + "\tfeature_region"

    return header, list(intersection), list(attributes_choosen_by_user)



# Implementation of the UCSC genome binning strategy -- heavily commented and
# with tests to help understand what's going on.
#
# Ryan Dale 2013
#
# With help from implementations in kent src and Brent Pedersen's cruzdb,
# specifically:
#
# http://genome-source.cse.ucsc.edu/gitweb/?p=kent.git\
#                                    ;a=blob;f=src/lib/binRange.c
# and
#
# https://github.com/brentp/cruzdb/blob/master/cruzdb/__init__.py
#
# For reference, Fig 7 in http://genome.cshlp.org/content/12/6/996.abstract
# looks like this:
# ----------------------------------------------------------------------------
# |                                  1                                       |
# ----------------------------------------------------------------------------
# |       2      |         3         |         4         |         5         |
# ----------------------------------------------------------------------------
# |6 | 7 | 8 | 9 | 10 | 11 | 12 | 13 | 14 | 15 | 16 | 17 | 18 | 19 | 20 | 21 |
# ----------------------------------------------------------------------------
#
#      AAAAAAAAAAAAAAAAAAAAAAA          BBBBBBBBBBBBBBB             CC
#
# The smallest bin feature "A" fits within is 1, but it also overlaps 2-3, 7-9.
# The smallest bin feature "B" fits within is 4, but it also overlaps 1, 14-17.
# The smallest bin feature "C" fits within is 20, but it also overlaps 1, 5.


# kent src: "How much to shift to get to next larger bin."
# So each level splits by 2**3 = 8 fold (octree?)
NEXT_SHIFT = 3

# kent src: "How much to shift to get to finest bin."
# 2 ** FIRST_SHIFT is the size of the smallest bin.
FIRST_SHIFT = 17

# These offsets are the bin numbers at the beginning of each level.  In Fig 7,
# these would be 6, 2, and 1.
#
# Bins with higher numbers (e.g. 585 to 4681) are the smaller-sized bins.  Run
# print_bin_sizes() for more info on this.
OFFSETS = [
    4096 + 512 + 64 + 8 + 1,  # bins 4681-585
    512 + 64 + 8 + 1,         # bins 585-73
    64 + 8 + 1,               # bins 73-9
    8 + 1,                    # bins 9-1
    1                         # bin  0
]

# for BED (0-based, half-open) or GFF (1-based, closed intervals)
COORD_OFFSETS = {'bed': 0, 'gff': 1}


def bins(start, stop, fmt='gff', one=True):
    """
    Uses the definition of a "genomic bin" described in Fig 7 of
    http://genome.cshlp.org/content/12/6/996.abstract.

    Parameters
    ----------
    one : boolean
        If `one=True` (default), then only return the smallest bin that
        completely contains these coordinates (useful for assigning a single
        bin).

        If `one=False`, then return the set of *all* bins that overlap these
        coordinates (useful for looking for features that could intersect)

    fmt : 'gff' | 'bed'
        This specifies 1-based start coords (gff) or 0-based start coords (bed)
    """
    # Jump to highest resolution bin that will fit these coords (depending on
    # whether we have a BED or GFF-style coordinate).
    #
    # Some GFF files include negative coords, which will throw off this
    # calculation.  If negative coords, then set the bin to the largest
    # possible.
    if start < 0:
        if one:
            return 1
        else:
            return set([1])

    if stop < 0:
        if one:
            return 1
        else:
            return set([1])

    start = (start - COORD_OFFSETS[fmt]) >> FIRST_SHIFT
    stop = (stop) >> FIRST_SHIFT

    # We always at least fit within the chrom, which is bin 1.
    bins = set([1])

    for offset in OFFSETS:
        # Since we're going from smallest to largest bins, the first one where
        # the feature's start and stop positions are both within the same bin
        # is the smallest one these coords fit within.
        if one:
            if start == stop:
                # Note that at this point, because of the bit-shifting, `start`
                # is the number of bins (at this current level).  So we need to
                # add it to `offset` to get the actual bin ID.
                return offset + start

        # See the Fig 7 reproduction above to see why range().
        bins.update(list(range(offset + start, offset + stop + 1)))

        # Move to the next level (8x larger bin size; i.e., 2**NEXT_SHIFT
        # larger bin size)
        start >>= NEXT_SHIFT
        stop >>= NEXT_SHIFT
    return bins


def print_bin_sizes():
    """
    Useful for debugging: how large is each bin, and what are the bin IDs?
    """
    for i, offset in enumerate(OFFSETS):
        binstart = offset
        try:
            binstop = OFFSETS[i + 1]
        except IndexError:
            binstop = binstart

        bin_size = 2 ** (FIRST_SHIFT + (i * NEXT_SHIFT))
        actual_size = bin_size

        # nice formatting
        bin_size, suffix = bin_size / 1024, 'Kb'
        if bin_size >= 1024:
            bin_size, suffix = bin_size / 1024, 'Mb'
        if bin_size >= 1024:
            bin_size, suffix = bin_size / 1024, 'Gb'
        size = '(%s %s)' % (bin_size, suffix)
        actual_size = '%s bp' % (actual_size)

        print('level: {i:1};  bins {binstart:<4} to {binstop:<4}; '
              'size: {actual_size:<12} {size:<6}'.format(**locals()))


def test():
    # These should obviously fit inside the first bin
    assert bins(0, 1, fmt='bed') == 4681
    assert bins(1, 1, fmt='gff') == 4681

    # All bins that overlap:
    assert bins(0, 1, fmt='bed', one=False) == set([1, 73, 9, 585, 4681])

    # Or, more generally,
    assert bins(0, 1, fmt='bed', one=False) == set(OFFSETS)

    # Since the smallest bin size is 2**17, this rather large feature should
    # fit completely inside the first bin, too
    assert bins(2 ** 17 - 1, 2 ** 17 - 1, fmt='bed') == 4681

    # or, more generally,
    assert bins(
        2 ** FIRST_SHIFT - 1,
        2 ** FIRST_SHIFT - 1,
        fmt='bed') == OFFSETS[0]

    # At exactly 2**17 in BED coords it moves over to the next bin
    assert bins(2 ** 17, 2 ** 17, fmt='bed') == 4682

    # or
    assert bins(
        2 ** FIRST_SHIFT,
        2 ** FIRST_SHIFT,
        fmt='bed') == OFFSETS[0] + 1

    # The other bins it overlaps should be similar to before, with the
    # exception of 4682
    assert bins(2 ** 17, 2 ** 17, fmt='bed', one=False) \
        == set([1, 9, 73, 585, 4682])

    assert bins(2 ** 17, 2 ** 17, fmt='bed', one=False)\
        .difference(bins(2 ** 17 - 1, 2 ** 17 - 1, fmt='bed', one=False)) \
        == set([4682])

    # Spanning the first and second smallest bins should cause it to bump up
    # a level
    assert bins(2 ** 17 - 1, 2 ** 17, fmt='bed') == 585

    # or
    assert bins(
        2 ** FIRST_SHIFT - 1,
        2 ** FIRST_SHIFT,
        fmt='bed') == OFFSETS[1]

    # Make it as big as it can get in this bin:
    assert bins(
        2 ** FIRST_SHIFT - 1,
        2 ** (FIRST_SHIFT + NEXT_SHIFT) - 1,
        fmt='bed') == OFFSETS[1]

    # Then make it big enough to jump another level
    assert bins(
        2 ** FIRST_SHIFT - 1,
        2 ** (FIRST_SHIFT + NEXT_SHIFT),
        fmt='bed') == OFFSETS[2]

    # When start or stop or both are negative, bin should be the largest
    # possible.
    assert bins(-1, 1000, one=True) == 1
    assert bins(-1, 1000, one=False) == set([1])
    assert bins(-1, -1000, one=True) == 1
    assert bins(-1, -1000, one=False) == set([1])
    assert bins(1, -1000, one=True) == 1
    assert bins(1, -1000, one=False) == set([1])

if __name__ == "__main__":
    print_bin_sizes()
    test()
